import logging
import operator
from collections import namedtuple

import numpy

from .cfdatetime import dt as cf_dt
from .cfdatetime import elements
from .data.data import Data
from .decorators import (
    _deprecated_kwarg_check,
    _manage_log_level_via_verbosity,
)
from .functions import inspect as cf_inspect
from .units import Units

logger = logging.getLogger(__name__)

# Define some useful units
_calendar_years = Units("calendar_years")
_calendar_months = Units("calendar_months")
_days = Units("days")
_hours = Units("hours")
_minutes = Units("minutes")
_seconds = Units("seconds")

# Define some useful constants
_one_year = Data(1, "calendar_years")
_one_day = Data(1, "day")
_one_hour = Data(1, "hour")
_one_minute = Data(1, "minute")
_one_second = Data(1, "second")

# Default month lengths in days
_default_month_lengths = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]

default_calendar = "gregorian"

Offset = namedtuple(
    "offset",
    ("year", "month", "day", "hour", "minute", "second", "microsecond"),
)

_relational_methods = (
    "__eq__",
    "__ne__",
    "__lt__",
    "__le__",
    "__gt__",
    "__ge__",
)


class TimeDuration:
    """A duration of time.

    The duration of time is a number of either calendar years,
    calendar months, days, hours, minutes or seconds.

    A calendar year (or month) is an arbitrary year (or month) in an
    arbitrary calendar. A calendar is as part of the time duration,
    but will be taken from the context in which the time duration
    instance is being used. For example, a calendar may be specified
    when creating time intervals (see below for examples).

    A default offset is specified that may be used by some
    applications to temporally position the time duration. For
    example, setting ``cf.TimeDuration(1, 'calendar_month', day=16,
    hour=12)`` will define a duration of one calendar month which, by
    default, starts at 12:00 on the 16th of the month. Note that the
    offset

    **Changing the units**

    The time duration's units may be changed in place by assigning
    equivalent units to the `~cf.TimeDuration.Units` attribute:

    >>> t = cf.TimeDuration(1, 'day')
    >>> t
    <CF TimeDuration: P1D (Y-M-D 00:00:00)>
    >>> t.Units = 's'
    >>> t
    <CF TimeDuration: PT86400.0S (Y-M-D 00:00:00)>
    >>> t.Units = cf.Units('minutes')
    >>> t
    <CF TimeDuration: PT1440.0M (Y-M-D 00:00:00)>
    >>> t.Units = 'calendar_months'
    ValueError: Can't set units (currently <Units: minutes>) to non-equivalent units <Units: calendar_months>

    **Creating time intervals**

    A time interval of exactly the time duration, starting or ending
    at a particular date-time, may be produced with the `interval`
    method. If elements of the start or end date-time are not
    specified, then default values are taken from the `!year`,
    `!month`, `!day`, `!hour`, `!minute`, and `!second` attributes of
    the time duration instance:

    >>> t = cf.TimeDuration(6, 'calendar_months')
    >>> t
    <CF TimeDuration: P6M (Y-M-01 00:00:00)>
    >>> t.interval(cf.dt(1999, 12))
    (cftime.DatetimeGregorian(1999-12-01 00:00:00),
     cftime.DatetimeGregorian(2000-06-01 00:00:00))
    >>> t = cf.TimeDuration(5, 'days', hour=6)
    >>> t
    <CF TimeDuration: P5D (Y-M-D 06:00:00)>
    >>> t.interval(cf.dt(2004, 3, 2), end=True)
    (cftime.DatetimeGregorian(2004-02-26 00:00:00),
     cftime.DatetimeGregorian(2004-03-02 00:00:00))
    >>> t.interval(cf.dt(2004, 3, 2, calendar='noleap'), end=True)
    (cftime.DatetimeNoLeap(2004-02-25 00:00:00),
     cftime.DatetimeNoLeap(2004-03-02 00:00:00))
    >>> t.interval(cf.dt(2004, 3, 2, calendar='360_day'), end=True)
    (cftime.Datetime360Day(2004-02-27 00:00:00),
     cftime.Datetime360Day(2004-03-02 00:00:00))
    >>> t.interval(cf.dt(2004, 3, 2, calendar='360_day'), end=True,
    ...     iso='start and duration')
    '2004-02-27 00:00:00/P5D'

    **Comparison operations**

    Comparison operations are defined for `cf.TimeDuration` objects,
    `cf.Data` objects, `numpy` arrays and numbers:

    >>> cf.TimeDuration(2, 'calendar_years') > cf.TimeDuration(
    ...     1, 'calendar_years')
    True
    >>> cf.TimeDuration(2, 'calendar_years') < cf.TimeDuration(
    ...     25, 'calendar_months')
    True
    >>> cf.TimeDuration(2, 'hours') <= cf.TimeDuration(1, 'days')
    True
    >>> cf.TimeDuration(2, 'hours') == cf.TimeDuration(1/12.0, 'days')
    True
    >>> cf.TimeDuration(2, 'days') == cf.TimeDuration(48, 'hours')
    True

    >>> cf.TimeDuration(2, 'hours') <= 2
    True
    >>> 30.5 != cf.TimeDuration(2, 'days')
    True

    >>> cf.TimeDuration(2, 'calendar_years') > numpy.array(1.5)
    True
    >>> type(cf.TimeDuration(2, 'calendar_years') > numpy.array(1.5))
    numpy.bool_
    >>> cf.TimeDuration(2, 'calendar_years') > numpy.array([1.5])
    array([ True])
    >>> numpy.array([[1, 12]]) > cf.TimeDuration(2, 'calendar_months')
    array([[False,  True]])

    >>> cf.TimeDuration(2, 'days') == cf.Data(2)
    <CF Data(): True>
    >>> cf.TimeDuration(2, 'days') == cf.Data([2.], 'days')
    <CF Data(1): [True]>
    >>> cf.Data([[60]], 'seconds') < cf.TimeDuration(2, 'days')
    <CF Data(1, 1): [[True]]>
    >>> cf.Data([1, 12], 'calendar_months') < cf.TimeDuration(
    ...     6, 'calendar_months')
    <CF Data(2): [True, False]>

    **Arithmetic operations**

    Arithmetic operations are defined for `cf.TimeDuration` objects,
    date-time-like objects (such as `cf.Datetime`,
    `datetime.datetime`, etc.), `cf.Data` objects, `numpy` arrays and
    numbers:

    >>> cf.TimeDuration(64, 'days') + cf.TimeDuration(28, 'days')
    <CF TimeDuration: P92D (Y-M-D 00:00:00)>
    >>> cf.TimeDuration(64, 'days') + cf.TimeDuration(12, 'hours')
    <CF TimeDuration: P65.0D (Y-M-D 00:00:00)>
    >>> cf.TimeDuration(64, 'days') + cf.TimeDuration(24, 'hours')
    <CF TimeDuration: P64.5D (Y-M-D 00:00:00)>
    >>> cf.TimeDuration(64, 'calendar_years') + cf.TimeDuration(
    ...     21, 'calendar_months')
    <CF TimeDuration: P65.75Y (Y-01-01 00:00:00)>

    >>> cf.TimeDuration(30, 'days') + 2
    <CF TimeDuration: P32D (Y-M-D 00:00:00)>
    >>> 4.5 + cf.TimeDuration(30, 'days')
    <CF TimeDuration: P34.5D (Y-M-D 00:00:00)>
    >>> cf.TimeDuration(64, 'calendar_years') - 2.5
    <CF TimeDuration: P61.5Y (Y-01-01 00:00:00)>

    >>> cf.TimeDuration(36, 'hours') / numpy.array(8)
    <CF TimeDuration: 4.5 hours (from Y-M-D h:00:00)>
    >>> cf.TimeDuration(36, 'hours') / numpy.array(8.0)
    <CF TimeDuration: 4.5 hours (from Y-M-D h:00:00)>
    >>> cf.TimeDuration(36, 'hours') // numpy.array(8.0)
    <CF TimeDuration: 4.0 hours (from Y-M-D h:00:00)>
    >>> cf.TimeDuration(36, 'calendar_months') * cf.Data([[2.25]])
    <CF TimeDuration: 81.0 calendar_months (from Y-M-01 00:00:00)>
    >>> cf.TimeDuration(36, 'calendar_months') // cf.Data([0.825])
    <CF TimeDuration: P43.0M (Y-01-01 00:00:00)>
    >>> cf.TimeDuration(36, 'calendar_months') % 10
    <CF TimeDuration: P6M (Y-01-01 00:00:00)>
    >>> cf.TimeDuration(36, 'calendar_months') % cf.Data(1, 'calendar_year')
    <CF TimeDuration: P0.0M (Y-01-01 00:00:00)>
    >>> cf.TimeDuration(36, 'calendar_months') % cf.Data(2, 'calendar_year')
    <CF TimeDuration: P12.0M (Y-01-01 00:00:00)>

    The in place operators (``+=``, ``//=``, etc.) are supported in a
    similar manner.

    **Attributes**

    ===========  =========================================================
    Attribute    Description
    ===========  =========================================================
    `!duration`  The length of the time duration in a `cf.Data` object
                 with units.
    `!year`      The default year for time interval creation.
    `!month`     The default month for time interval creation.
    `!day`       The default day for time interval creation.
    `!hour`      The default hour for time interval creation.
    `!minute`    The default minute for time interval creation.
    `!second`    The default second for time interval creation.
    ===========  =========================================================


    **Constructors**

    For convenience, the following functions may also be used to
    create time duration objects:

    ========  ============================================================
    Function  Description
    ========  ============================================================
    `cf.Y`    Create a time duration of calendar years.
    `cf.M`    Create a time duration of calendar months.
    `cf.D`    Create a time duration of days.
    `cf.h`    Create a time duration of hours.
    `cf.m`    Create a time duration of minutes.
    `cf.s`    Create a time duration of seconds.
    ========  ============================================================

    .. versionadded:: 1.0

    .. seealso:: `cf.dt`, `cf.Data`, `cf.Datetime`

    """

    def __init__(
        self, duration, units=None, month=1, day=1, hour=0, minute=0, second=0
    ):
        """**Initialisation**

        :Parameters:

            duration: data-like
                The length of the time duration.

                A data-like object is any object containing array-like or
                scalar data which could be used to create a `cf.Data`
                object.

                *Parameter example:*
                  Instances, ``x``, of following types are all examples of
                  data-like objects (because ``cf.Data(x)`` creates a
                  valid `cf.Data` object), `int`, `float`, `str`, `tuple`,
                  `list`, `numpy.ndarray`, `cf.Data`, `cf.Coordinate`,
                  `cf.Field`.

            units: `str` or `cf.Units`, optional
                The units of the time duration. Required if, and only if,
                *duration* is not a `cf.Data` object which already
                contains the units. Units must be one of calendar years,
                calendar months, days, hours, minutes or seconds.

                *Parameter example:*
                  ``units='calendar_months'``

                *Parameter example:*
                  ``units='days'``

                *Parameter example:*
                  ``units=cf.Units('calendar_years')``

             month, day, hour, minute, second: `int` or `None`, optional
                The offset used when creating, with the `bounds` method, a
                time interval containing a given date-time.

                .. note:: The offset element *month* is ignored unless the
                          time duration is at least 1 calendar year.

                          The offset element *day* is ignored unless the
                          time duration is at least 1 calendar month.

                          The offset element *hour* is ignored unless the
                          time duration is at least 1 day

                          The offset element *minute* is ignored unless
                          the time duration is at least 1 hour.

                          The offset element *second* is ignored unless
                          the time duration is at least 1 minute

                *Parameter example:*
                  >>> cf.TimeDuration(1, 'calendar_month').bounds(
                  ...     cf.dt('2000-1-8'))
                  (cftime.DatetimeGregorian(2000-01-01 00:00:00),
                   cftime.DatetimeGregorian(2000-02-01 00:00:00))
                  >>> cf.TimeDuration(1, 'calendar_month', day=15).bounds(
                  ...     cf.dt('2000-1-8'))
                  (cftime.DatetimeGregorian(1999-12-15 00:00:00),
                   cftime.DatetimeGregorian(2000-01-15 00:00:00))
                  >>> cf.TimeDuration(
                  ...     1, 'calendar_month', month=4, day=30).bounds(
                  ...         cf.dt('2000-1-8'))
                  (cftime.DatetimeGregorian(1999-12-30 00:00:00),
                   cftime.DatetimeGregorian(2000-01-30 00:00:00))

        **Examples**

        >>> t = cf.TimeDuration(cf.Data(3 , 'calendar_years'))
        >>> t = cf.TimeDuration(cf.Data(12 , 'hours'))
        >>> t = cf.TimeDuration(18 , 'calendar_months')
        >>> t = cf.TimeDuration(30 , 'days')
        >>> t = cf.TimeDuration(1 , 'day', hour=6)

        """
        if units is not None:
            units = Units(units)
            self.duration = Data(abs(duration))
            if self.duration.Units.istime:
                self.duration.Units = units
            else:
                self.duration = Data(self.duration.array, units)
        else:
            self.duration = abs(Data.asdata(duration))
            units = self.duration.Units
            if not units.istime:
                raise ValueError(f"Bad units: {units!r}")

        if not (units.iscalendartime or units.istime):
            raise ValueError(
                f"Can't create {self.__class__.__name__} of {self.duration}"
            )

        duration = self.duration

        offset = [None, month, day, hour, minute, second, 0]
        if units.equivalent(_calendar_years):
            if duration < _one_year:
                offset[1] = None
        else:
            offset[1] = None
            offset[2] = None
            if duration < _one_day:
                offset[3] = None
                if duration < _one_hour:
                    offset[4] = None
                    if duration < _one_minute:
                        offset[5] = None
        #            if units <= _hours and duration < _one_day:
        #                offset[3] = None
        #                if units <= _minutes and duration < _one_hour:
        #                    offset[4] = None
        #                    if units <= _seconds and duration < _one_minute:
        #                        offset[5] = None

        self.offset = Offset(*offset)

        # TODO should offset be None for all "higher" units

        self._compound = False

        self._NotImplemented_RHS_Data_op = True

    def __abs__(self):
        """x.__abs__() <==> abs(x)

        .. versionadded:: 1.4

        """
        out = self.copy()
        out.duration = abs(self.duration)
        return out

    def __array__(self, *dtype):
        """Returns a `numpy` array representing the time duration."""
        return self.duration.__array__(*dtype)

    def __data__(self):
        """Returns a new reference to the `!duration` attribute."""
        return self.duration

    def __deepcopy__(self, memo):
        """Used if copy.deepcopy is called."""
        return self.copy()

    def __neg__(self):
        """x.__neg__() <==> -x.

        .. versionadded:: 1.4

        """
        out = self.copy()
        out.duration *= -1
        return out

    def __bool__(self):
        """Truth value testing and the built-in operation `bool`

        x.__bool__() <==> bool(x)

        """
        return bool(self.duration)

    def __int__(self):
        """x.__int__() <==> int(x)"""
        return int(self.duration)

    def __repr__(self):
        """x.__repr__() <==> repr(x)"""
        return f"<CF {self.__class__.__name__}: {str(self)}>"

    def __str__(self):
        """x.__str__() <==> str(x)"""
        yyy = [
            x if y is None else f"{y:0>2}"
            for x, y in zip(("Y", "M", "D", "h", "m", "s"), self.offset)
        ]

        return "{0} ({1}-{2}-{3} {4}:{5}:{6})".format(self.iso, *yyy)

    def __ge__(self, other):
        """The rich comparison operator ``>=``

        x.__ge__(y) <==> x>=y

        """

        return self._apply_binary_comparison(other, "__ge__")

    def __gt__(self, other):
        """The rich comparison operator ``>``

        x.__gt__(y) <==> x>y

        """

        return self._apply_binary_comparison(other, "__gt__")

    def __le__(self, other):
        """The rich comparison operator ``<=``

        x.__le__(y) <==> x<=y

        """
        return self._apply_binary_comparison(other, "__le__")

    def __lt__(self, other):
        """The rich comparison operator ``<``

        x.__lt__(y) <==> x<y

        """
        return self._apply_binary_comparison(other, "__lt__")

    def __eq__(self, other):
        """The rich comparison operator ``==``

        x.__eq__(y) <==> x==y

        """
        return self._apply_binary_comparison(other, "__eq__")

    def __ne__(self, other):
        """The rich comparison operator ``!=``

        x.__ne__(y) <==> x!=y

        """
        return self._apply_binary_comparison(other, "__ne__")

    def __add__(self, other):
        """The binary arithmetic operation ``+``

        x.__add__(y) <==> x + y

        .. versionadded:: 1.4

        """
        return self._apply_binary_arithmetic(
            other, "__add__", may_be_datetime=True
        )

        return NotImplemented

    def __sub__(self, other):
        """The binary arithmetic operation ``-``

        x.__sub__(y) <==> x - y

        .. versionadded:: 1.4

        """
        return self._apply_binary_arithmetic(
            other, "__sub__", may_be_datetime=True
        )

    def __mul__(self, other):
        """The binary arithmetic operation ``*``

        x.__mul__(y) <==> x*y

        """
        return self._apply_binary_arithmetic(other, "__mul__")

    def __div__(self, other):
        """The binary arithmetic operation ``/``

        x.__div__(y) <==> x/y

        """
        return self._apply_binary_arithmetic(other, "__div__")

    def __floordiv__(self, other):
        """The binary arithmetic operation ``//``

        x.__floordiv__(y) <==> x//y

        """
        return self._apply_binary_arithmetic(other, "__floordiv__")

    def __truediv__(self, other):
        """The binary arithmetic operation ``/`` (true division)

        x.__truediv__(y) <==> x/y

        """
        return self._apply_binary_arithmetic(other, "__truediv__")

    def __iadd__(self, other):
        """The augmented arithmetic assignment ``+=``

        x.__iadd__(y) <==> x+=y

        """
        return self._apply_binary_arithmetic(
            other, "__iadd__", aug_assignment=True, skip_data_return=True
        )

    def __idiv__(self, other):
        """The augmented arithmetic assignment ``/=``

        x.__idiv__(y) <==> x /= y

        """
        return self._apply_binary_arithmetic(
            other, "__idiv__", aug_assignment=True
        )

    def __itruediv__(self, other):
        """The augmented arithmetic assignment ``/=`` (true division)

        x.__truediv__(y) <==> x/y

        """
        return self._apply_binary_arithmetic(
            other, "__itruediv__", aug_assignment=True, skip_data_return=True
        )

    def __ifloordiv__(self, other):
        """The augmented arithmetic assignment ``//=``

        x.__ifloordiv__(y) <==> x//=y

        """
        return self._apply_binary_arithmetic(
            other, "__ifloordiv__", aug_assignment=True, skip_data_return=True
        )

    def __imul__(self, other):
        """The augmented arithmetic assignment ``*=``

        x.__imul__(y) <==> x *= y

        """
        return self._apply_binary_arithmetic(
            other, "__imul__", aug_assignment=True, skip_data_return=True
        )

    def __isub__(self, other):
        """The augmented arithmetic assignment ``-=``

        x.__isub__(y) <==> x -= y

        """
        return self._apply_binary_arithmetic(
            other, "__isub__", aug_assignment=True, skip_data_return=True
        )

    def __imod__(self, other):
        """The augmented arithmetic assignment ``%=``

        x.__imod__(y) <==> x%=y

        """
        return self._apply_binary_arithmetic(
            other, "__imod__", aug_assignment=True
        )

    def __radd__(self, other):
        """The binary arithmetic operation ``+`` with reflected
        operands.

        x.__radd__(y) <==> y+x

        """
        return self + other

    def __rmul__(self, other):
        """The binary arithmetic operation ``*`` with reflected
        operands.

        x.__rmul__(y) <==> y*x

        """
        return self * other

    def __rsub__(self, other):
        """The binary arithmetic operation ``-`` with reflected
        operands.

        x.__rsub__(y) <==> y - x

        .. versionadded:: 1.4

        """
        return -self + other

    def __mod__(self, other):
        """The binary arithmetic operation ``%``

        x.__mod__(y) <==> x % y

        """
        return self._apply_binary_arithmetic(other, "__mod__")

    def __rmod__(self, other):
        """The binary arithmetic operation ``%`` with reflected
        operands.

        x.__rmod__(y) <==> y % x

        """
        if isinstance(other, (self.__class__, int, float)):
            return self._binary_operation(other, "__rmod__")

        return NotImplemented

    # ----------------------------------------------------------------
    # Private methods
    # ----------------------------------------------------------------
    def _apply_binary_comparison(self, other, op):
        """Apply a binary comparison operation on general data.

        .. versionadded:: 3.12.0

        :Parameters:

            other: the object to compare with.

            op: `str`, the binary comparison operator to apply.

        """

        if isinstance(other, (self.__class__, int, float)):
            return bool(self._binary_operation(other, op))

        if isinstance(other, Data):
            return self._data_binary_operation(other, op)

        return NotImplemented

    def _apply_binary_arithmetic(
        self,
        other,
        op,
        aug_assignment=False,
        may_be_datetime=False,
        skip_data_return=False,
    ):
        """Apply a binary arithmetic operation with general data.

        .. versionadded:: 3.12.0

        :Parameters:

            other: the object to compare with.

            op: `str`, the binary arithmetic operator to apply.

            aug_assignment: `bool`
                Whether the arithmetic operation is one of
                augmented assignment. By default it is not.

            may_be_datetime: `bool`
                Whether it is logically permissible for other to be a
                datetime, which is False by default.

            skip_data_return: `bool`
                Whether to skip logic to apply the operation on the
                datetime data if other has Data type. By default,
                do not skip.

        """
        check_simple_types = [int, float]
        if may_be_datetime:
            check_simple_types.append(self.__class__)

        if isinstance(other, tuple(check_simple_types)):
            return self._binary_operation(other, op, aug_assignment)

        if may_be_datetime and hasattr(other, "timetuple"):
            # other is a date-time object
            try:
                return self._datetime_arithmetic(other, getattr(operator, op))
            except TypeError:
                return NotImplemented

        if not skip_data_return and isinstance(other, Data):
            return self._data_arithmetic(other, op, aug_assignment)

        return NotImplemented

    def _binary_operation(self, other, method, inplace=False):
        """Implement binary operations on the datetime."""
        if inplace:
            new = self
        else:
            new = self.copy()

        self_units = self.duration.Units

        other = getattr(other, "duration", other)

        duration = getattr(new.duration, method)(other)

        duration.squeeze(inplace=True)

        if duration.size != 1:
            raise ValueError(
                f"Can't create {self.__class__.__name__} with more than one "
                f"value: {duration!r}"
            )

        if duration < 0:
            raise ValueError(
                f"Can't create {self.__class__.__name__} with with negative "
                f"duration {duration!r}"
            )

        if (
            not duration.Units.equals(self.duration.Units)
            and method not in _relational_methods
        ):
            # Operator is not one of ==, !=, >=, >, <=, <
            raise ValueError(
                f"Can't create {self.__class__.__name__} of {duration.Units!r}"
            )

        if method not in _relational_methods:
            # Operator is not one of ==, !=, >=, >, <=, <
            duration.Units = self_units

        new.duration = duration

        return new

    def _data_binary_operation(self, other, method, inplace=False):
        """Implement binary operations on the datetime data."""
        if inplace:
            new = self
        else:
            new = self.copy()

        return getattr(new.duration, method)(other)

    def _datetime_arithmetic(self, other, op):
        """Apply binary arithmetic operations on the datetime.

        .. versionadded:: 1.4

        :Parameters:

            other: any object with a `timetuple` method

            op: `str`

        :Returns:

        """

        def _dHMS(duration, other, calendar, op):
            units = Units(f"{duration.Units.units} since {other}", calendar)
            d = op(Data(0.0, units), duration)
            return d.datetime_array.item(())

        duration = self.duration
        units = duration.Units

        if units == _calendar_years:
            months0 = duration.datum() * 12
            months = int(months0)
            if months != months0:
                raise ValueError(
                    "Fractional months are not supported for  "
                    f"date calculations: {months0}"
                )
        elif units == _calendar_months:
            months0 = duration.datum()
            months = int(months0)
            if months != months0:
                raise ValueError(
                    "Fractional months are not supported for "
                    f"date calculations: {months0}"
                )
        else:
            months = None

        calendar = getattr(other, "calendar", None)
        #        if calendar == '':
        #            calendar = None

        if months is not None:
            y, m = divmod(op(other.month, months), 12)
            if not m:
                y -= 1
                m = 12

            y = other.year + y

            d = other.day
            if calendar != "":
                max_days = self.days_in_month(y, m, calendar)
                if d > max_days:
                    d = max_days

            return other.replace(year=y, month=m, day=d)
        else:
            return _dHMS(duration, other, calendar, op)

    def _data_arithmetic(self, other, method, inplace=False):
        """Apply binary arithmetic operations on the datetime data."""
        try:
            dt = other.datetime_array
        except ValueError:
            return self._binary_operation(other, method, inplace=inplace)
        else:
            out = []
            for d in dt.flat:
                if d is numpy.ma.masked:
                    out.append(None)
                else:
                    out.append(getattr(self, method)(d))

            dt[...] = numpy.reshape(out, dt.shape)

            return Data(dt, units=other.Units)

    def _offset(self, dt):
        """Calculate the time duration offset.

        .. versionadded:: 1.4

        :Parameters:

            dt: `cf.Datetime`
                The date-time from which to calculate the offset.
                *dt* may be any date-time-like object, such as
                `cf.Datetime`, `datetime.datetime`,
                `netCDF4.netcdftime.datetime`, etc.

        :Returns:

            `Datetime`
                The date-time object representing the offset.

        """
        return cf_dt(
            *[
                (i if j is None else j)
                for i, j in zip(elements(dt), self.offset)
            ],
            calendar=getattr(dt, "calendar", None),
        )

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def iso(self):
        """Return the time duration as an ISO 8601-like time duration
        string.

        .. versionadded:: 1.0

        **Examples**

        >>> cf.TimeDuration(45, 'days').iso
        'P45D'
        >>> cf.TimeDuration(10, 'calendar_years').iso
        'P10Y'
        >>> cf.s(5.67).iso
        'PT5.67S'
        >>> cf.M(18.5).iso
        'P18.5M'

        """
        duration = self.duration
        units = duration.Units

        date_prefix = f"P{duration.datum()}"
        if units.equals(_calendar_months):
            return f"{date_prefix}M"
        if units.equals(_calendar_years):
            return f"{date_prefix}Y"
        if units.equals(_days):
            return f"{date_prefix}D"

        time_prefix = f"PT{duration.datum()}"
        if units.equals(_hours):
            return f"{time_prefix}H"
        if units.equals(_minutes):
            return f"{time_prefix}M"
        if units.equals(_seconds):
            return f"{time_prefix}S"

        raise ValueError(f"Bad {self.__class__.__name__} units: {units!r}")

    @property
    def isint(self):
        """True if the time duration is a whole number.

        .. versionadded:: 1.0

        **Examples**

        >>> cf.TimeDuration(2, 'hours').isint
        True
        >>> cf.TimeDuration(2.0, 'hours').isint
        True
        >>> cf.TimeDuration(2.5, 'hours').isint
        False

        """
        duration = self.duration

        if duration.dtype.kind == "i":
            return True

        duration = duration.datum()
        return int(duration) == float(duration)

    @property
    def Units(self):
        """The units of the time duration.

        .. versionadded:: 1.0

        **Examples**

        >>> cf.TimeDuration(3, 'days').Units
        <CF Units: days>

        >>> t = cf.TimeDuration(cf.Data(12, 'calendar_months'))
        >>> t.Units
        <CF Units: calendar_months>
        >>> t.Units = cf.Units('calendar_years')
        >>> t.Units
        <CF Units: calendar_years>
        >>> t
        <CF TimeDuration: 1.0 calendar_years (from Y-01-01 00:00:00)>

        """
        return self.duration.Units

    @Units.setter
    def Units(self, value):
        duration = getattr(self, "duration", None)
        if duration is None:
            raise AttributeError(
                "Can't set units when there is no duration attribute"
            )

        self.duration.Units = Units(value)

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def copy(self):
        """Return a deep copy.

        ``t.copy()`` is equivalent to ``copy.deepcopy(t)``.

        .. versionadded:: 1.0

        :Returns:

                The deep copy.

        **Examples**

        >>> u = t.copy()

        """
        new = TimeDuration.__new__(TimeDuration)
        new.__dict__ = self.__dict__.copy()
        new.duration = self.duration.copy()
        return new

    @classmethod
    def days_in_month(
        cls, year, month, calendar=None, leap_month=2, month_lengths=None
    ):
        """The number of days in a specific month in a specific year in
        a specific calendar.

        .. versionadded:: 1.4

        :Parameters:

            year: `int`
                The calendar year to query.

            month: `int`
                The one- or two-digit integer representing the calendar
                month to query (numbered by order in the calendar year,
                for example 2 for February, 10 for October).

            calendar: `str`, optional
                By default, calendar is the mixed Gregorian/Julian
                calendar as defined by Udunits.

            leap_month: `int`, optional
                The leap month. By default the leap month is 2, i.e. the
                second month of the year.

            month_lengths: sequence of `int`, optional
                By default, *month_lengths* is ``[31, 28, 31, 30, 31, 30,
                31, 31, 30, 31, 30, 31]``.

        :Returns:

            `int`
                The number of days in the specified month.

        **Examples**

        >>> cf.TimeDuration.days_in_month(2004, 2, calendar='360_day')
        30

        """
        if month_lengths is None:
            month_lengths = _default_month_lengths
        elif len(month_lengths) != 12:
            raise ValueError("month_lengths must be a sequence of 12 elements")

        month1 = month - 1

        if calendar in [
            None,
            "standard",
            "gregorian",
            "proleptic_gregorian",
            "",
        ]:
            length = month_lengths[month1]
            if month == leap_month and (
                year % 400 == 0 or (year % 100 != 0 and year % 4 == 0)
            ):
                length += 1
        elif calendar == "360_day":
            length = 30
        elif calendar in ("noleap", "365_day"):
            length = month_lengths[month1]
        elif calendar in ("all_leap", "366_day"):
            length = month_lengths[month1]
            if month == leap_month:
                length += 1
        elif calendar == "julian":
            length = month_lengths[month1]
            if month == leap_month and year % 4 == 0:
                length += 1

        return length

    @_deprecated_kwarg_check("traceback", version="3.0.0", removed_at="4.0.0")
    @_manage_log_level_via_verbosity
    def equals(
        self, other, rtol=None, atol=None, verbose=None, traceback=False
    ):
        """True if two time durations are equal.

        .. versionadded:: 1.0

        .. seealso:: `equivalent`

        :Parameters:

            other:
                The object to compare for equality.

            rtol: `float`, optional
                The relative tolerance for all numerical comparisons, By
                default the value returned by the `rtol` function is used.

            atol: `float`, optional
                The absolute tolerance for all numerical comparisons, By
                default the value returned by the `atol` function is used.

            verbose: `int` or `str` or `None`, optional
                If an integer from ``-1`` to ``3``, or an equivalent string
                equal ignoring case to one of:

                * ``'DISABLE'`` (``0``)
                * ``'WARNING'`` (``1``)
                * ``'INFO'`` (``2``)
                * ``'DETAIL'`` (``3``)
                * ``'DEBUG'`` (``-1``)

                set for the duration of the method call only as the minimum
                cut-off for the verboseness level of displayed output (log)
                messages, regardless of the globally-configured `cf.log_level`.
                Note that increasing numerical value corresponds to increasing
                verbosity, with the exception of ``-1`` as a special case of
                maximal and extreme verbosity.

                Otherwise, if `None` (the default value), output messages will
                be shown according to the value of the `cf.log_level` setting.

                Overall, the higher a non-negative integer or equivalent string
                that is set (up to a maximum of ``3``/``'DETAIL'``) for
                increasing verbosity, the more description that is printed to
                convey information about the test for contiguity.

            traceback: deprecated at version 3.0.0
                Use the *verbose* parameter instead.

        :Returns:

            `bool`
                Whether or not the two instances are equal.

        **Examples**

        >>> t = cf.TimeDuration(36, 'calendar_months')
        >>> u = cf.TimeDuration(3, 'calendar_years')
        >>> t == u
        True
        >>> t.equals(u, verbose=2)
        TimeDuration: Different durations: <CF Data: 36 calendar_months>, <CF Data: 3 calendar_years>
        False

        """
        # Check each instance's id
        if self is other:
            return True

        # Check that each instance is the same type
        if self.__class__ != other.__class__:
            logger.info(
                "%s: Different type: %s"
                % (self.__class__.__name__, other.__class__.__name__)
            )  # pragma: no cover
            return False

        self__dict__ = self.__dict__.copy()
        other__dict__ = other.__dict__.copy()

        d0 = self__dict__.pop("duration", None)
        d1 = other__dict__.pop("duration", None)

        if not d0.equals(d1):
            logger.info(
                "%s: Different durations: %r, %r"
                % (self.__class__.__name__, d0, d1)
            )  # pragma: no cover
            return False

        if self__dict__ != other__dict__:
            logger.info(
                "%s: Different default date-time elements: "
                "%r != %r"
                % (self.__class__.__name__, self__dict__, other__dict__)
            )  # pragma: no cover
            return False

        return True

    @_deprecated_kwarg_check("traceback", version="3.0.0", removed_at="4.0.0")
    @_manage_log_level_via_verbosity
    def equivalent(
        self, other, rtol=None, atol=None, verbose=None, traceback=False
    ):
        """True if two time durations are logically equivalent.

        .. versionadded:: 1.0

        .. seealso:: `equals`

        :Parameters:

            other:
                The object to compare for equivalence.

            rtol: `float`, optional
                The relative tolerance for all numerical comparisons, By
                default the value returned by the `rtol` function is used.

            atol: `float`, optional
                The absolute tolerance for all numerical comparisons, By
                default the value returned by the `atol` function is used.

            verbose: `int` or `str` or `None`, optional
                If an integer from ``-1`` to ``3``, or an equivalent string
                equal ignoring case to one of:

                * ``'DISABLE'`` (``0``)
                * ``'WARNING'`` (``1``)
                * ``'INFO'`` (``2``)
                * ``'DETAIL'`` (``3``)
                * ``'DEBUG'`` (``-1``)

                set for the duration of the method call only as the minimum
                cut-off for the verboseness level of displayed output (log)
                messages, regardless of the globally-configured `cf.log_level`.
                Note that increasing numerical value corresponds to increasing
                verbosity, with the exception of ``-1`` as a special case of
                maximal and extreme verbosity.

                Otherwise, if `None` (the default value), output messages will
                be shown according to the value of the `cf.log_level` setting.

                Overall, the higher a non-negative integer or equivalent string
                that is set (up to a maximum of ``3``/``'DETAIL'``) for
                increasing verbosity, the more description that is printed to
                convey information about the test for contiguity.

            traceback: deprecated at version 3.0.0
                Use the *verbose* parameter instead.

        :Returns:

            `bool`
                Whether or not the two instances logically equivalent.

        **Examples**

        >>> t = cf.TimeDuration(36, 'calendar_months')
        >>> u = cf.TimeDuration(3, 'calendar_years')
        >>> t == u
        True
        >>> t.equivalent(u)
        True
        >>> t.equals(u, verbose=2)
        TimeDuration: Different durations: <CF Data: 12 calendar_months>, <CF Data: 1 calendar_years>
        False

        """
        # Check each instance's id
        if self is other:
            return True

        # Check that each instance is the same type
        if self.__class__ != other.__class__:
            logger.info(
                "%s: Different type: %s"
                % (self.__class__.__name__, other.__class__.__name__)
            )  # pragma: no cover
            return False

        self__dict__ = self.__dict__.copy()
        other__dict__ = other.__dict__.copy()

        d0 = self__dict__.pop("duration", None)
        d1 = other__dict__.pop("duration", None)
        if d0 != d0:
            logger.info(
                "%s: Non-equivalent durations: %r, %r"
                % (self.__class__.__name__, d0, d1)
            )  # pragma: no cover
            return False

        if self__dict__ != other__dict__:
            logger.info(
                "%s: Non-equivalent default date-time elements: "
                "%r != %r"
                % (self.__class__.__name__, self__dict__, other__dict__)
            )  # pragma: no cover
            return False

        return True

    def inspect(self):
        """Inspect the attributes.

        .. versionadded:: 1.0

        .. seealso:: `cf.inspect`

        :Returns:

            `None`

        """
        print(cf_inspect(self))  # pragma: no cover

    def interval(self, dt, end=False, iso=None):  # calendar=None,
        """Return a time interval of exactly the time duration.

        The start (or end, if the *end* parameter is True) date-time of
        the time interval is determined by the date-time given by the *dt*
        parameter.

        .. versionadded:: 1.0

        .. seealso:: `bounds`

        :Parameters:

            dt:
                The date-time. One of:

                * A `str` specifying an ISO 8601-like date-time string (in
                  which non-Gregorian calendar dates are allowed).

                * `datetime.datetime or subclass `cftime.datetime` (such
                  as `Datetime360Day`).

            end: `bool`, optional
                If True then the date-time given by the *year*, *month*,
                *day*, *hour*, *minute* and *second* parameters defines
                the end of the time interval. By default it defines the
                start of the time interval.

            iso: `str`, optional
                Return the time interval as an ISO 8601 time interval
                string rather than the default of a tuple of date-time
                objects. Valid values are (with example outputs for the
                time interval "3 years from 2007-03-01 13:00:00"):

                  ========================  =============================================
                  iso                       Example output
                  ========================  =============================================
                  ``'start and end'``       ``'2007-03-01 13:00:00/2010-03-01 13:00:00'``
                  ``'start and duration'``  ``'2007-03-01 13:00:00/P3Y'``
                  ``'duration and end'``    ``'P3Y/2010-03-01 13:00:00'``
                  ========================  =============================================

        :Returns:

                The date-times at each end of the time interval. The first
                date-time is always earlier than or equal to the second
                date-time. If *iso* has been set then an ISO 8601 time
                interval string is returned instead of a tuple.

        **Examples**

        >>> t = cf.TimeDuration(6, 'calendar_months')
        >>> t
        <CF TimeDuration: P6M (Y-M-01 00:00:00)>
        >>> t.interval(cf.dt(1999, 12))
        (cftime.DatetimeGregorian(1999-12-01 00:00:00),
         cftime.DatetimeGregorian(2000-06-01 00:00:00))

        >>> t = cf.TimeDuration(5, 'days', hour=6)
        >>> t
        <CF TimeDuration: P5D (Y-M-D 06:00:00)>
        >>> t.interval(cf.dt(2004, 3, 2), end=True)
        (cftime.DatetimeGregorian(2004-02-26 00:00:00),
         cftime.DatetimeGregorian(2004-03-02 00:00:00))
        >>> t.interval(cf.dt(2004, 3, 2, calendar='noleap'), end=True)
        (cftime.DatetimeNoLeap(2004-02-25 00:00:00),
         cftime.DatetimeNoLeap(2004-03-02 00:00:00))
        >>> t.interval(cf.dt(2004, 3, 2, calendar='360_day'), end=True)
        (cftime.Datetime360Day(2004-02-27 00:00:00),
         cftime.Datetime360Day(2004-03-02 00:00:00))
        >>> t.interval(cf.dt(2004, 3, 2, calendar='360_day'),
        ...            end=True, iso='start and duration')
        '2004-02-27 00:00:00/P5D'

        Create `cf.Query` objects for a time interval - one including both
        bounds and one which excludes the upper bound:

        >>> t = cf.TimeDuration(2, 'calendar_years')
        >>> interval = t.interval(cf.dt(1999, 12))
        >>> c = cf.wi(*interval)
        >>> c
        <CF Query: (wi [cftime.DatetimeGregorian(1999-12-01 00:00:00), cftime.DatetimeGregorian(2001-12-01 00:00:00)])>
        >>> c == cf.dt('2001-1-1', calendar='gregorian')
        True

        Create a `cf.Query` object which may be used to test where a time
        coordinate object's bounds lie inside a time interval:

        >>> t = cf.TimeDuration(1, 'calendar_months')
        >>> c = cf.cellwi(*t.interval(cf.dt(2000, 1), end=True))
        >>> c
        <CF Query: [lower_bounds(ge 1999-12-01 00:00:00) & upper_bounds(le 2000-01-01 00:00:00)]>

        Create ISO 8601 time interval strings:

        >>> t = cf.TimeDuration(6, 'calendar_years')
        >>> t.interval(cf.dt(1999, 12), end=True, iso='start and end')
        '1993-12-01 00:00:00/1999-12-01 00:00:00'
        >>> t.interval(cf.dt(1999, 12), end=True, iso='start and duration')
        '1993-12-01 00:00:00/P6Y'
        >>> t.interval(cf.dt(1999, 12), end=True, iso='duration and end')
        'P6Y/1999-12-01 00:00:00'

        """

        def _dHMS(duration, dt, end):
            calendar = dt.calendar
            if not calendar:
                calendar = None

            units = Units(f"{duration.Units.units} since {dt}", calendar)
            dt1 = Data(0.0, units)

            if not end:
                dt1 += duration
            else:
                dt1 -= duration

            dt1 = dt1.datetime_array.item(())

            if not end:
                return dt, dt1  # dt.copy(), dt1
            else:
                return dt1, dt  # dt1, dt.copy()

        calendar = getattr(dt, "calendar", default_calendar)
        if calendar == "":
            calendar = default_calendar

        dt = cf_dt(dt, calendar=calendar)

        duration = self.duration
        units = duration.Units

        if units == _calendar_years:
            months = duration.datum() * 12
            int_months = int(months)
            if int_months != months:
                raise ValueError(
                    "Can't create a time interval of a non-integer number "
                    f"of calendar months: {months}"
                )
        elif units == _calendar_months:
            months = duration.datum()
            int_months = int(months)
            if int_months != months:
                raise ValueError(
                    "Can't create a time interval of a non-integer number "
                    f"of calendar months: {months}"
                )
        else:
            int_months = None

        if int_months is not None:
            if not end:
                y, month1 = divmod(dt.month + int_months, 12)

                if not month1:
                    y -= 1
                    month1 = 12

                year1 = dt.year + y

                max_days = self.days_in_month(year1, month1, calendar)

                day1 = dt.day
                if day1 > max_days:
                    day1 = max_days

                dt0 = dt  # .copy()
                dt1 = dt.replace(year=year1, month=month1, day=day1)
            else:
                y, month0 = divmod(dt.month - int_months, 12)

                if not month0:
                    y -= 1
                    month0 = 12

                year0 = dt.year + y

                max_days = self.days_in_month(year0, month0, calendar)
                day0 = dt.day
                if day0 > max_days:
                    day0 = max_days

                dt0 = dt.replace(year=year0, month=month0, day=day0)
                dt1 = dt  # 0.copy()
        else:
            dt0, dt1 = _dHMS(duration, dt, end)

        if not iso:
            return dt0, dt1

        if iso == "start and end":
            return f"{dt0}/{dt1}"
        if iso == "start and duration":
            return f"{dt0}/{self.iso}"
        if iso == "duration and end":
            return f"{self.iso}/{dt1}"

    def bounds(self, dt, direction=True):
        """Return a time interval containing a date-time.

        The interval spans the time duration and starts and ends at
        date-times consistent with the time duration's offset.

        The offset of the time duration is used to modify the bounds.

        .. versionadded:: 1.2.3

        .. seealso:: `cf.dt`, `cf.Datetime`, `interval`, `offset`

        :Parameters:

            dt: date-time-like
                The date-time to be contained by the interval. *dt* may be
                any date-time-like object, such as `cf.Datetime`,
                `datetime.datetime`, `netCDF4.netcdftime.datetime`, etc.

                *Parameter example:*
                  To find bounds around 1999-16-1 in the Gregorian
                  calendar you could use ``dt=cf.dt(1999, 1, 16)`` or
                  ``dt=datetime.datetime(1999, 1, 16)`` (See `cf.dt` for
                  details).

            direction: `bool`, optional
                If `False` then the bounds are decreasing. By default the
                bounds are increasing. Note that ``t.bounds(dt,
                direction=False)`` is equivalent to
                ``t.bounds(dt)[::-1]``.

        :Returns:

            `tuple`
                The two bounds.

        **Examples**

        >>> t = cf.M()
        >>> t.bounds(cf.dt(2000, 1, 1))
        (cftime.DatetimeGregorian(2000, 1, 1, 0, 0, 0, 0, has_year_zero=False),
         cftime.DatetimeGregorian(2000, 2, 1, 0, 0, 0, 0, has_year_zero=False))

        >>> t = cf.M(1)
        >>> t.bounds(cf.dt(2000, 3, 1))
        (cftime.DatetimeGregorian(2000, 3, 1, 0, 0, 0, 0, has_year_zero=False),
         cftime.DatetimeGregorian(2000, 4, 1, 0, 0, 0, 0, has_year_zero=False))

        >>> t = cf.M(1, day=15)
        >>> t.bounds(cf.dt(2000, 3, 1))
        (cftime.DatetimeGregorian(2000, 2, 15, 0, 0, 0, 0, has_year_zero=False),
         cftime.DatetimeGregorian(2000, 3, 15, 0, 0, 0, 0, has_year_zero=False))

        >>> t = cf.M(2, day=15)
        >>> t.bounds(cf.dt(2000, 3, 1), direction=False)
        (cftime.DatetimeGregorian(2000, 3, 15, 0, 0, 0, 0, has_year_zero=False),
         cftime.DatetimeGregorian(2000, 1, 15, 0, 0, 0, 0, has_year_zero=False))

        """
        abs_self = abs(self)

        calendar = getattr(dt, "calendar", default_calendar)
        if calendar == "":
            calendar = default_calendar

        dt = cf_dt(dt, calendar=calendar)

        dt0 = self._offset(dt)

        if dt0 > dt:
            dt1 = dt0
            dt0 = dt1 - abs_self
        else:
            dt1 = dt0 + abs_self

        if direction:
            return (dt0, dt1)
        else:
            return (dt1, dt0)

    def is_day_factor(self):
        """Return True if an integer multiple of the time duration is
        equal to one day.

        .. versionadded:: 1.0

        :Returns:

            `bool`

        **Examples**

        >>> cf.TimeDuration(0.5, 'days').is_day_factor()
        True

        >>> cf.TimeDuration(1, 'days').is_day_factor()
        True
        >>> cf.TimeDuration(0.25, 'days').is_day_factor()
        True
        >>> cf.TimeDuration(0.3, 'days').is_day_factor()
        False
        >>> cf.TimeDuration(2, 'days').is_day_factor()
        False

        >>> cf.TimeDuration(24, 'hours').is_day_factor()
        True
        >>> cf.TimeDuration(6, 'hours').is_day_factor()
        True
        >>> cf.TimeDuration(7, 'hours').is_day_factor()
        False
        >>> cf.TimeDuration(27, 'hours').is_day_factor()
        False

        >>> cf.TimeDuration(1440, 'minutes').is_day_factor()
        True
        >>> cf.TimeDuration(15, 'minutes').is_day_factor()
        True
        >>> cf.TimeDuration(17, 'minutes').is_day_factor()
        False
        >>> cf.TimeDuration(2007, 'minutes').is_day_factor()
        False

        >>> cf.TimeDuration(86400, 'seconds').is_day_factor()
        True
        >>> cf.TimeDuration(45, 'seconds').is_day_factor()
        True
        >>> cf.TimeDuration(47, 'seconds').is_day_factor()
        False
        >>> cf.TimeDuration(86401, 'seconds').is_day_factor()
        False

        >>> cf.TimeDuration(1, 'calendar_months').is_day_factor()
        False
        >>> cf.TimeDuration(1, 'calendar_years').is_day_factor()
        False

        """
        try:
            return not Data(1, "day") % self.duration
        except ValueError:
            return False


def Y(duration=1, month=1, day=1, hour=0, minute=0, second=0):
    """Return a time duration of calendar years in a `cf.TimeDuration`
    object.

    ``cf.Y()`` is equivalent to ``cf.TimeDuration(1,
    'calendar_year')``.

    ``cf.Y(duration, *args, **kwargs)`` is equivalent to
    ``cf.TimeDuration(duration, 'calendar_year', *args, **kwargs)``.

    .. versionadded:: 1.0

    .. seealso:: `cf.M`, `cf.D`, `cf.h`, `cf.m`, `cf.s`

    :Parameters:

        duration: number, optional
            The number of calendar years in the time duration.

        month, day, hour, minute, second: `int`, optional
            The default date-time elements for defining the start and
            end of a time interval based on this time duration. See
            `cf.TimeDuration` and `cf.TimeDuration.interval` for
            details.

            *Parameter example:*
              ``cf.Y(month=12)`` is equivalent to ``cf.TimeDuration(1,
              'calendar_years', month=12)``.

    :Returns:

        `TimeDuration`
            The new `cf.TimeDuration` object.

    **Examples**

    >>> cf.Y()
    <CF TimeDuration: P1Y (Y-01-01 00:00:00)>
    >>> cf.Y(10, month=12)
    <CF TimeDuration: P10Y (Y-12-01 00:00:00)>
    >>> cf.Y(15, month=4, day=2, hour=12, minute=30, second=2)
    <CF TimeDuration: P15Y (Y-04-02 12:30:02)>
    >>> cf.Y(0)
    <CF TimeDuration: P0Y (Y-M-01 00:00:00)>

    """
    return TimeDuration(
        duration,
        "calendar_years",
        month=month,
        day=day,
        hour=hour,
        minute=minute,
        second=second,
    )


def M(duration=1, month=1, day=1, hour=0, minute=0, second=0):
    """Return a time duration of calendar months in a `cf.TimeDuration`
    object.

    ``cf.M()`` is equivalent to ``cf.TimeDuration(1, 'calendar_month')``.

    .. versionadded:: 1.0

    .. seealso:: `cf.Y`, `cf.D`, `cf.h`, `cf.m`, `cf.s`

    :Parameters:

        duration: number, optional
            The number of calendar months in the time duration.

        year, month, day, hour, minute, second: `int`, optional
            The default date-time elements for defining the start and
            end of a time interval based on this time duration. See
            `cf.TimeDuration` and `cf.TimeDuration.interval` for
            details.

            *Parameter example:*
              ``cf.M(day=16)`` is equivalent to ``cf.TimeDuration(1,
              'calendar_months', day=16)``.

    :Returns:

        `TimeDuration`
            The new `cf.TimeDuration` object.

    **Examples**

    >>> cf.M()
    <CF TimeDuration: P1M (Y-M-01 00:00:00)>
    >>> cf.M(3, day=16)
    <CF TimeDuration: P3M (Y-M-16 00:00:00)>
    >>> cf.M(24, day=2, hour=12, minute=30, second=2)
    <CF TimeDuration: P24M (Y-01-02 12:30:02)>
    >>> cf.M(0)
    <CF TimeDuration: P0M (Y-M-01 00:00:00)>

    """
    return TimeDuration(
        duration,
        "calendar_months",
        month=month,
        day=day,
        hour=hour,
        minute=minute,
        second=second,
    )


def D(duration=1, month=1, day=1, hour=0, minute=0, second=0):
    """Return a time duration of days in a `cf.TimeDuration` object.

    ``cf.D()`` is equivalent to ``cf.TimeDuration(1, 'day')``.

    .. versionadded:: 1.0

    .. seealso:: `cf.Y`, `cf.M`, `cf.h`, `cf.m`, `cf.s`

    :Parameters:

        duration: number, optional
            The number of days in the time duration.

        month, day, hour, minute, second: `int`, optional
            The default date-time elements for defining the start and end
            of a time interval based on this time duration. See
            `cf.TimeDuration` and `cf.TimeDuration.interval` for details.

            *Parameter example:*
               ``cf.D(hour=12)`` is equivalent to ``cf.TimeDuration(1,
               'day', hour=12)``.

    :Returns:

        `TimeDuration`
            The new `cf.TimeDuration` object.

    **Examples**

    >>> cf.D()
    <CF TimeDuration: P1D (Y-M-D 00:00:00)>
    >>> cf.D(5, hour=12)
    <CF TimeDuration: P5D (Y-M-D 12:00:00)>
    >>> cf.D(48.5, minute=30)
    <CF TimeDuration: P48.5D (Y-M-D 00:30:00)>
    >>> cf.D(0.25, hour=6, minute=30, second=20)
    <CF TimeDuration: P0.25D (Y-M-D 06:30:20)>
    >>> cf.D(0)
    <CF TimeDuration: P0D (Y-M-D 00:00:00)>

    """
    return TimeDuration(
        duration,
        "days",
        month=month,
        day=day,
        hour=hour,
        minute=minute,
        second=second,
    )


def h(duration=1, month=1, day=1, hour=0, minute=0, second=0):
    """Return a time duration of hours in a `cf.TimeDuration` object.

    ``cf.h()`` is equivalent to ``cf.TimeDuration(1, 'hour')``.

    .. versionadded:: 1.0

    .. seealso:: `cf.Y`, `cf.M`, `cf.D`, `cf.m`, `cf.s`

    :Parameters:

        duration: number, optional
            The number of hours in the time duration.

        year, month, day, hour, minute, second: `int`, optional
            The default date-time elements for defining the start and
            end of a time interval based on this time duration. See
            `cf.TimeDuration` and `cf.TimeDuration.interval` for
            details.

            *Parameter example:*
              ``cf.h(minute=30)`` is equivalent to
              ``cf.TimeDuration(1, 'hour', minute=30)``.

    :Returns:

        `TimeDuration`
            The new `cf.TimeDuration` object.

    **Examples**

    >>> cf.h()
    <CF TimeDuration: PT1H (Y-M-D h:00:00)>
    >>> cf.h(3, minute=15)
    <CF TimeDuration: PT3H (Y-M-D h:15:00)>
    >>> cf.h(0.5)
    <CF TimeDuration: PT0.5H (Y-M-D h:00:00)>
    >>> cf.h(6.5, minute=15, second=45)
    <CF TimeDuration: PT6.5H (Y-M-D h:15:45)>
    >>> cf.h(0)
    <CF TimeDuration: PT0H (Y-M-D h:00:00)>

    """
    return TimeDuration(
        duration,
        "hours",
        month=month,
        day=day,
        hour=hour,
        minute=minute,
        second=second,
    )


def m(duration=1, month=1, day=1, hour=0, minute=0, second=0):
    """Return a time duration of minutes in a `cf.TimeDuration` object.

    ``cf.m()`` is equivalent to ``cf.TimeDuration(1, 'minute')``.

    .. versionadded:: 1.0

    .. seealso:: `cf.Y`, `cf.M`, `cf.D`, `cf.h`, `cf.s`

    :Parameters:

        duration: number, optional
            The number of hours in the time duration.

        month, day, hour, minute, second: `int`, optional
            The default date-time elements for defining when a time
            interval based on this time duration begins or ends. See
            `cf.TimeDuration` and `cf.TimeDuration.interval` for
            details.

            *Parameter example:*
              ``cf.m(second=30)`` is equivalent to
              ``cf.TimeDuration(1, 'minute', second=30)``.

    :Returns:

        `TimeDuration`
            The new `cf.TimeDuration` object.

    **Examples**

    >>> cf.m()
    <CF TimeDuration: PT1M (Y-M-D h:m:00)>
    >>> cf.m(30, second=15)
    <CF TimeDuration: PT30M (Y-M-D h:m:15)>
    >>> cf.m(0.5)
    <CF TimeDuration: PT0.5M (Y-M-D h:m:00)>
    >>> cf.m(2.5, second=45)
    <CF TimeDuration: PT2.5M (Y-M-D h:m:45)>
    >>> cf.m(0)
    <CF TimeDuration: PT0M (Y-M-D h:m:00)>

    """
    return TimeDuration(
        duration,
        "minutes",
        month=month,
        day=day,
        hour=hour,
        minute=minute,
        second=second,
    )


def s(duration=1, month=1, day=1, hour=0, minute=0, second=0):
    """Return a time duration of seconds in a `cf.TimeDuration` object.

    ``cf.s()`` is equivalent to ``cf.TimeDuration(1, 'second')``.

    .. versionadded:: 1.0

    .. seealso:: `cf.Y`, `cf.M`, `cf.D`, `cf.h`, `cf.m`

    :Parameters:

        duration: number, optional
            The number of hours in the time duration.

        month, day, hour, minute, second: `int`, optional
            The default date-time elements for defining the start and
            end of a time interval based on this time duration. See
            `cf.TimeDuration` and `cf.TimeDuration.interval` for
            details.

            *Parameter example:*
              ``cf.s(hour=6)`` is equivalent to ``cf.TimeDuration(1,
              'seconds', hour=6)``.

    :Returns:

        `TimeDuration`
            The new `cf.TimeDuration` object.

    **Examples**

    >>> cf.s()
    <CF TimeDuration: PT1S (Y-M-D h:m:s)>
    >>> cf.s().interval(cf.dt(1999, 12, 1))
    (cftime.DatetimeGregorian(1999-12-01 00:00:00),
     cftime.DatetimeGregorian(1999-12-01 00:00:01))
    >>> cf.s(30)
    <CF TimeDuration: PT30S (Y-M-D h:m:s)>
    >>> cf.s(0.5)
    <CF TimeDuration: PT0.5S (Y-M-D h:m:s)>
    >>> cf.s(12.25)
    <CF TimeDuration: PT12.25S (Y-M-D h:m:s)>
    >>> cf.s(2.5, year=1999, hour=12)
    cf.s(30, month=2, hour=12)
    <CF TimeDuration: PT30S (Y-M-D h:m:s)>
    >>> cf.s(0)
    <CF TimeDuration: PT0S (Y-M-D h:m:s)>

    """
    return TimeDuration(
        duration,
        "seconds",
        month=month,
        day=day,
        hour=hour,
        minute=minute,
        second=second,
    )



from . import ConstructList, mixin
from .functions import _DEPRECATION_ERROR_METHOD


class FieldList(mixin.FieldDomainList, ConstructList):
    """An ordered sequence of field constructs.

    A field list supports the python list-like operations (such as
    indexing and methods like `!append`). These methods provide
    functionality similar to that of a built-in list. The main
    difference is that when a field construct element needs to be
    assessed for equality its `~cf.Field.equals` method is used, rather
    than the ``==`` operator.

    """

    def __init__(self, fields=None):
        """**Initialisation**

        :Parameters:

            fields: (sequence of) `Field`, optional
                 Create a new list with these field constructs.

        """
        super().__init__(constructs=fields)

    def concatenate(self, axis=0, cull_graph=True):
        """Join the sequence of fields within the field list together.

        This is different to `cf.aggregate` because it does not
        account for all metadata. For example, it assumes that the
        axis order is the same in each field.

        .. versionadded:: 1.0

        .. seealso:: `cf.aggregate`, `Field.concatenate`,
                     `Data.concatenate`

        :Parameters:

            axis: `int`, optional
                The axis along which the arrays will be joined. The
                default is 0. Note that scalar arrays are treated as
                if they were one dimensional.

            {{cull_graph: `bool`, optional}}

        :Returns:

            `Field`
                The field generated from the concatenation of all of
                the fields contained in the input field list.

        """
        return self[0].concatenate(self, axis=axis, cull_graph=cull_graph)

    def select_by_naxes(self, *naxes):
        """Select field constructs by property.

        To find the inverse of the selection, use a list comprehension
        with `~cf.Field.match_by_naxes` method of the construct
        elements. For example, to select all constructs which do *not*
        have 3-dimensional data:

           >>> gl = cf.FieldList(
           ...     f for f in fl if not f.match_by_naxes(3)
           ... )

        .. versionadded:: 3.0.0

        .. seealso:: `select`, `select_by_identity`,
                     `select_by_construct`, `select_by_property`,
                     `select_by_rank`, `select_by_units`

        :Parameters:

            naxes: `int`, optional
                Select field constructs whose data spans a particular
                number of domain axis constructs.

                A number of domain axis constructs is given by an
                `int`.

                If no numbers are provided then all field constructs
                are selected.

        :Returns:

            `FieldList`
                The matching field constructs.

        **Examples**

        >>> f = cf.read("file.nc")
        >>> f
        [<CF Field: specific_humidity(latitude(5), longitude(8)) 1>,
         <CF Field: air_temperature(atmosphere_hybrid_height_coordinate(1), grid_latitude(10), grid_longitude(9)) K>]
        >>> f.select_by_naxes()
        [<CF Field: specific_humidity(latitude(5), longitude(8)) 1>,
        <CF Field: air_temperature(atmosphere_hybrid_height_coordinate(1), grid_latitude(10), grid_longitude(9)) K>]
        >>> f.select_by_naxes(1)
        []
        >>> f.select_by_naxes(2)
        [<CF Field: specific_humidity(latitude(5), longitude(8)) 1>]
        >>> f.select_by_naxes(3)
        [<CF Field: air_temperature(atmosphere_hybrid_height_coordinate(1), grid_latitude(10), grid_longitude(9)) K>]

        """
        return type(self)(f for f in self if f.match_by_naxes(*naxes))

    def select_by_units(self, *units, exact=True):
        """Select field constructs by units.

        To find the inverse of the selection, use a list comprehension
        with `~cf.Field.match_by_units` method of the construct
        elements. For example, to select all constructs whose units
        are *not* ``'km'``:

           >>> gl = cf.FieldList(
           ...     f for f in fl if not f.match_by_units('km')
           ... )

        .. versionadded:: 3.0.0

        .. seealso:: `select`, `select_by_identity`,
                     `select_by_construct`, `select_by_naxes`,
                     `select_by_rank`, `select_by_property`

        :Parameters:

            units: optional
                Select field constructs. By default all field
                constructs are selected. May be one or more of:

                  * The units of a field construct.

                Units are specified by a string or compiled regular
                expression (e.g. 'km', 'm s-1',
                ``re.compile('^kilo')``, etc.) or a `Units` object
                (e.g. ``Units('km')``, ``Units('m s-1')``, etc.).

            exact: `bool`, optional
                If `False` then select field constructs whose units
                are equivalent to any of those given by *units*. For
                example, metres and are equivalent to kilometres. By
                default, field constructs whose units are exactly one
                of those given by *units* are selected. Note that the
                format of the units is not important, i.e. 'm' is
                exactly the same as 'metres' for this purpose.

        :Returns:

            `FieldList`
                The matching field constructs.

        **Examples**

        >>> gl = fl.select_by_units('metres')
        >>> gl = fl.select_by_units('m')
        >>> gl = fl.select_by_units('m', 'kilogram')
        >>> gl = fl.select_by_units(Units('m'))
        >>> gl = fl.select_by_units('km', exact=False)
        >>> gl = fl.select_by_units(Units('km'), exact=False)
        >>> gl = fl.select_by_units(re.compile('^met'))
        >>> gl = fl.select_by_units(Units('km'))
        >>> gl = fl.select_by_units(Units('kg m-2'))

        """
        return type(self)(
            f for f in self if f.match_by_units(*units, exact=exact)
        )

    def select_field(self, *identities, default=ValueError()):
        """Select a unique field construct by its identity.

        .. versionadded:: 3.0.4

        .. seealso:: `select`, `select_by_identity`

        :Parameters:

            identities: optional
                Select the field construct by one or more of

                * A construct identity.

                  {{construct selection identity}}

            default: optional
                Return the value of the *default* parameter if a
                unique field construct can not be found.

                {{default Exception}}

        :Returns:

            `Field`
                The unique matching field construct.

        **Examples**

        >>> fl
        [<CF Field: specific_humidity(latitude(73), longitude(96)) 1>,
         <CF Field: specific_humidity(latitude(73), longitude(96)) 1>,
         <CF Field: air_temperature(time(12), latitude(64), longitude(128)) K>]
        >>> fl.select_field('air_temperature')
        <CF Field: air_temperature(time(12), latitude(64), longitude(128)) K>
        >>> f.select_field('specific_humidity')
        ValueError: Multiple fields found
        >>> f.select_field('specific_humidity', 'No unique field')
        'No unique field'
        >>> f.select_field('snowfall_amount')
        ValueError: No fields found

        """
        out = self.select_by_identity(*identities)

        if not out:
            if default is None:
                return

            return self._default(
                default, "select_field() can't return 0 fields"
            )

        n = len(out)
        if n > 1:
            if default is None:
                return

            return self._default(
                default, f"select_field() can't return {n} fields"
            )

        return out[0]

    # ----------------------------------------------------------------
    # Deprecated attributes and methods
    # ----------------------------------------------------------------
    def _parameters(self, d):
        """Deprecated at version 3.0.0."""
        _DEPRECATION_ERROR_METHOD(
            self, "_parameters", version="3.0.0", removed_at="4.0.0"
        )  # pragma: no cover

    def _deprecated_method(self, name):
        """Deprecated at version 3.0.0."""
        _DEPRECATION_ERROR_METHOD(
            self, "_deprecated_method", version="3.0.0", removed_at="4.0.0"
        )  # pragma: no cover

    def set_equals(
        self,
        other,
        rtol=None,
        atol=None,
        ignore_data_type=False,
        ignore_fill_value=False,
        ignore_properties=(),
        ignore_compression=False,
        ignore_type=False,
        traceback=False,
    ):
        """Deprecated at version 3.0.0.

        Use method 'equals' with unordered=True instead.

        """
        _DEPRECATION_ERROR_METHOD(
            self,
            "set_equals",
            "Use method 'equals' with unordered=True instead.",
            version="3.0.0",
            removed_at="4.0.0",
        )  # pragma: no cover

    def select1(self, *args, **kwargs):
        """Deprecated at version 3.0.0.

        Use method 'fl.select_field' instead.

        """
        _DEPRECATION_ERROR_METHOD(
            self,
            "select1",
            "Use method 'fl.select_field' instead.",
            version="3.0.0",
            removed_at="4.0.0",
        )  # pragma: no cover



import os
from functools import cmp_to_key

import numpy

from . import cInterface
from .extraData import ExtraDataUnpacker


class UMFileException(Exception):
    pass


class File:
    """A class for a UM file that gives a view of the file including
    sets of PP records combined into variables."""

    def __init__(
        self, path, byte_ordering=None, word_size=None, fmt=None, parse=True
    ):
        """Open and parse a UM file.

        The optional *byte_ordering*, *word_size* and *fmt* arguments
        specify the file type. If all three are set, then this forces the
        file type; otherwise, the file type is autodetected and any of
        them that are set are ignored.

        :Parameters:

            path: `str`
                The name of the UM file.

            byte_ordering: `str`, optional
                'little_endian' or 'big_endian'

            word_size: `int`, optional
                4 or 8

            fmt: `str`, optional
                'FF' or 'PP'

            parse: `bool`, optional
                The default action is to open the file, store the file
                type from the arguments or autodetection as described
                above, and then parse the contents, giving a tree of
                variables and records under the `File` object. However, if
                *parse* is False, then an object is returned in which the
                last step is omitted, so only the file type is stored, and
                there are no variables under it. Such an object can be
                passed when instantiating Rec objects, and contains
                sufficient info about the file type to ensure that the
                `get_data` method of those `Rec` objects will work.

        """
        c = cInterface.CInterface()
        self._c_interface = c

        self.path = path
        self.fd = None
        self.open_fd()

        if byte_ordering and word_size and fmt:
            self.fmt = fmt
            self.byte_ordering = byte_ordering
            self.word_size = word_size
        else:
            self._detect_file_type()

        self.path = path
        file_type_obj = c.create_file_type(
            self.fmt, self.byte_ordering, self.word_size
        )

        # Set the word size used to interpret file pointers
        c.set_word_size(file_type_obj)

        if parse:
            # --------------------------------------------------------
            # Work out information from the file and store it in the
            # `vars` attribute.
            #
            # Note that the word size used to interpret file pointers
            # needs to have been previously set.
            # --------------------------------------------------------
            info = c.parse_file(self.fd, file_type_obj)
            self.vars = info["vars"]
            self._add_back_refs()

    def open_fd(self):
        """(Re)open the low-level file descriptor.

        :Returns:

            `int`
                The file descriptor.

        """
        if self.fd is None:
            self.fd = os.open(self.path, os.O_RDONLY)

        return self.fd

    def close_fd(self):
        """Close the low-level file descriptor.

        :Returns:

            `None`

        """
        if self.fd:
            os.close(self.fd)

        self.fd = None

    def _detect_file_type(self):
        """Store string values describing the auto-detected file type.

        :Returns:

            `None`

        """
        c = self._c_interface
        try:
            file_type_obj = c.detect_file_type(self.fd)
        except Exception:
            self.close_fd()
            raise IOError(f"File {self.path} has unsupported format")

        d = c.file_type_obj_to_dict(file_type_obj)
        self.fmt = d["fmt"]
        self.byte_ordering = d["byte_ordering"]
        self.word_size = d["word_size"]

    def _add_back_refs(self):
        """Add file attribute to `Var` objects, and both `!file` and
        `!var` attributes to `Rec` objects.

        The important one is the file attribute in the `Rec` object, as
        this is used when reading data. The others are provided for extra
        convenience.

        :Returns:

            `None`

        """
        for var in self.vars:
            var.file = self
            for rec in var.recs:
                rec.var = var
                rec.file = self


class Var:
    """Container for some information about variables."""

    def __init__(self, recs, nz, nt, supervar_index=None):
        self.recs = recs
        self.nz = nz
        self.nt = nt
        self.supervar_index = supervar_index

    @staticmethod
    def _compare(x, y):
        """Method equivalent to the Python 2 'cmp'.

        Note that (x > y) - (x < y) is equivalent but not as performant
        since it would not short-circuit.

        :Returns:

            `int`

        """
        if x == y:
            return 0
        elif x > y:
            return 1
        else:
            return -1

    def _compare_recs_by_extra_data(self, a, b):
        """Compare records with respect to their extra data.

        :Returns:

            `int`

        """
        return self._compare(a.get_extra_data(), b.get_extra_data())

    def _compare_recs_by_orig_order(self, a, b):
        """Compare records with respect to their original order.

        :Returns:

            `int`

        """
        return self._compare(self.recs.index(a), self.recs.index(b))

    def group_records_by_extra_data(self):
        """Group records by matching extra data.

        Returns a list of (sub)lists of records where each record
        within each sublist has matching extra data (if
        any). Therefore, if the whole variable has consistent extra
        data then the returned value will be a list of length 1.

        Within each group, the ordering of returned records is the
        same as in the `!recs` attribute.

        :Returns:

            `list`

        """
        compare = self._compare_recs_by_extra_data
        recs = self.recs[:]
        n = len(recs)
        if n == 0:
            # shouldn't have a var without records, but...
            return []

        #        recs.sort(compare) #python2
        recs.sort(key=cmp_to_key(compare))

        # optimise simple case - if two ends of a sorted list match,
        # the whole list matches
        if not compare(recs[0], recs[-1]):
            return [self.recs[:]]

        groups = []
        this_grp = []
        for i, rec in enumerate(recs):
            this_grp.append(rec)
            if i == n - 1 or compare(rec, recs[i + 1]):
                this_grp.sort(key=self._compare_recs_by_orig_order)
                groups.append(this_grp)
                this_grp = []

        return groups


class Rec:
    """Container for some information about records."""

    def __init__(
        self,
        int_hdr,
        real_hdr,
        hdr_offset,
        data_offset,
        disk_length,
        file=None,
    ):
        """Default instantiation, which stores the supplied headers and
        offsets.

        :Parameters:

            file: `File`, optional
                Used to set the `!file` attribute. Does not need to be
                supplied, but if it is not then it will have to be set on
                the returned `Rec` object before calling `get_data` will
                work. If set it should be set to the `File` object that
                contains the returned `Rec` object. Normally this would be
                done by the calling code instantiating via `File` rather
                than directly.

        """
        self.int_hdr = int_hdr
        self.real_hdr = real_hdr
        self.hdr_offset = hdr_offset
        self.data_offset = data_offset
        self.disk_length = disk_length
        self._extra_data = None
        if file:
            self.file = file

    @classmethod
    def from_file_and_offsets(cls, file, hdr_offset, data_offset, disk_length):
        """Instantiate a `Rec` object from the `File` object and the
        header and data offsets.

        The headers are read in, and also the record object is ready for
        calling `get_data`.

        :Parameters:

            file: `File`
                A view of a file including sets of PP records combined
                into variables.

            hdr_offset: `int`
                The start word in the file of the header.

            data_offset: `int`
                The start word in the file of the data.

            disk_length: `int`
                The length in words of the data in the file.

        :Returns:

             `Rec`

        """
        c = file._c_interface
        int_hdr, real_hdr = c.read_header(
            file.fd, hdr_offset, file.byte_ordering, file.word_size
        )

        return cls(
            int_hdr, real_hdr, hdr_offset, data_offset, disk_length, file=file
        )

    def read_extra_data(self):
        """Read the extra data associated with the record.

        :Returns:

            `numpy.ndarray`

        """
        c = self.file._c_interface
        file = self.file

        (
            extra_data_offset,
            extra_data_length,
        ) = c.get_extra_data_offset_and_length(
            self.int_hdr, self.data_offset, self.disk_length
        )

        raw_extra_data = c.read_extra_data(
            file.fd,
            extra_data_offset,
            extra_data_length,
            file.byte_ordering,
            file.word_size,
        )

        edu = ExtraDataUnpacker(
            raw_extra_data, file.word_size, file.byte_ordering
        )

        return edu.get_data()

    def get_extra_data(self):
        """Get extra data associated with the record.

        This is done either by reading or using cached read.

        :Returns:

            `numpy.ndarray`

        """
        if self._extra_data is None:
            self._extra_data = self.read_extra_data()

        return self._extra_data

    def get_type_and_num_words(self):
        """Get the data type and number of words.

        :Returns:

            `numpy.dtype`, `int`

        """
        c = self.file._c_interface
        ntype, num_words = c.get_type_and_num_words(self.int_hdr)
        if ntype == "integer":
            dtype = numpy.dtype(c.file_data_int_type)
        elif ntype == "real":
            dtype = numpy.dtype(c.file_data_real_type)

        return dtype, num_words

    def get_data(self):
        """Get the data array associated with the record.

        :Returns:

            `numpy.ndarray`

        """
        c = self.file._c_interface
        file = self.file
        data_type, nwords = c.get_type_and_num_words(self.int_hdr)

        return c.read_record_data(
            file.fd,
            self.data_offset,
            self.disk_length,
            file.byte_ordering,
            file.word_size,
            self.int_hdr,
            self.real_hdr,
            data_type,
            nwords,
        )


if __name__ == "__main__":
    import sys

    path = sys.argv[1]
    f = File(path)
    print(f.fmt, f.byte_ordering, f.word_size)
    print("num variables: %s" % len(f.vars))
    for varno, var in enumerate(f.vars):
        print()
        print("var %s: nz = %s, nt = %s" % (varno, var.nz, var.nt))
        for recno, rec in enumerate(var.recs):
            print("var %s record %s" % (varno, recno))
            print("hdr offset: %s" % rec.hdr_offset)
            print("data offset: %s" % rec.data_offset)
            print("disk length: %s" % rec.disk_length)
            print("int hdr: %s" % rec.int_hdr)
            print("real hdr: %s" % rec.real_hdr)
            print("data: %s" % rec.get_data())
            print("extra_data: %s" % rec.get_extra_data())
            print("type %s, num words: %s" % rec.get_type_and_num_words())
            # if recno == 1:
            #     rec._extra_data['y'] += .01
            #     print("massaged_extra_data: %s" % rec.get_extra_data())
            print("-----------------------")

        print("all records", var.recs)
        print(
            "records grouped by extra data ", var.group_records_by_extra_data()
        )
        print("===============================")

    f.close_fd()

    # also read a record using saved metadata
    if f.vars:

        fmt = f.fmt
        byte_ordering = f.byte_ordering
        word_size = f.word_size
        myrec = f.vars[0].recs[0]
        hdr_offset = myrec.hdr_offset
        data_offset = myrec.data_offset
        disk_length = myrec.disk_length

        del f

        fnew = File(
            path,
            fmt=fmt,
            byte_ordering=byte_ordering,
            word_size=word_size,
            parse=False,
        )

        rnew = Rec.from_file_and_offsets(
            fnew, hdr_offset, data_offset, disk_length
        )
        print("record read using saved file type and offsets:")
        print("int hdr: %s" % rnew.int_hdr)
        print("real hdr: %s" % rnew.real_hdr)
        print("data: %s" % rnew.get_data())
        print("extra data: %s" % rnew.get_extra_data())
        print("nx = %s" % rnew.int_hdr[18])
        print("ny = %s" % rnew.int_hdr[17])
        rdata = open("recdata0.txt", "w")
        for value in rnew.get_data():
            rdata.write("%s\n" % value)
        rdata.close()



from ...functions import (
    _DEPRECATION_ERROR_ATTRIBUTE,
    _DEPRECATION_ERROR_METHOD,
    DeprecationError,
)


class DataClassDeprecationsMixin:
    """Deprecated attributes and methods for the Data class."""

    def __hash__(self):
        """The built-in function `hash`.

        Deprecated at version TODODASKVER. Consider using the
        `cf.hash_array` function instead.

        Generating the hash temporarily realises the entire array in
        memory, which may not be possible for large arrays.

        The hash value is dependent on the data-type and shape of the data
        array. If the array is a masked array then the hash value is
        independent of the fill value and of data array values underlying
        any masked elements.

        The hash value may be different if regenerated after the data
        array has been changed in place.

        The hash value is not guaranteed to be portable across versions of
        Python, numpy and cf.

        :Returns:

            `int`
                The hash value.

        **Examples**

        >>> print(d.array)
        [[0 1 2 3]]
        >>> d.hash()
        -8125230271916303273
        >>> d[1, 0] = numpy.ma.masked
        >>> print(d.array)
        [[0 -- 2 3]]
        >>> hash(d)
        791917586613573563
        >>> d.hardmask = False
        >>> d[0, 1] = 999
        >>> d[0, 1] = numpy.ma.masked
        >>> d.hash()
        791917586613573563
        >>> d.squeeze()
        >>> print(d.array)
        [0 -- 2 3]
        >>> hash(d)
        -7007538450787927902
        >>> d.dtype = float
        >>> print(d.array)
        [0.0 -- 2.0 3.0]
        >>> hash(d)
        -4816859207969696442

        """
        _DEPRECATION_ERROR_METHOD(
            self,
            "__hash__",
            message="Consider using 'cf.hash_array' function array instead.",
            version="TODODASKVER",
            removed_at="5.0.0",
        )

    @property
    def _HDF_chunks(self):
        """The HDF chunksizes.

        Deprecated at version TODODASKVER.

        DO NOT CHANGE IN PLACE.

        """
        _DEPRECATION_ERROR_ATTRIBUTE(
            self, "_HDF_chunks", version="TODODASKVER", removed_at="5.0.0"
        )  # pragma: no cover

    @_HDF_chunks.setter
    def _HDF_chunks(self, value):
        _DEPRECATION_ERROR_ATTRIBUTE(
            self, "_HDF_chunks", version="TODODASKVER", removed_at="5.0.0"
        )  # pragma: no cover

    @_HDF_chunks.deleter
    def _HDF_chunks(self):
        _DEPRECATION_ERROR_ATTRIBUTE(
            self, "_HDF_chunks", version="TODODASKVER", removed_at="5.0.0"
        )  # pragma: no cover

    @property
    def Data(self):
        """Deprecated at version 3.0.0, use attribute `data` instead."""
        _DEPRECATION_ERROR_ATTRIBUTE(
            self,
            "Data",
            "Use attribute 'data' instead.",
            version="3.0.0",
            removed_at="4.0.0",
        )  # pragma: no cover

    @property
    def dtvarray(self):
        """Deprecated at version 3.0.0."""
        _DEPRECATION_ERROR_ATTRIBUTE(
            self, "dtvarray", version="3.0.0", removed_at="4.0.0"
        )  # pragma: no cover

    @property
    def in_memory(self):
        """True if the array is retained in memory.

        Deprecated at version TODODASKVER.

        """
        _DEPRECATION_ERROR_ATTRIBUTE(
            self, "in_memory", version="TODODASKVER", removed_at="5.0.0"
        )  # pragma: no cover

    @property
    def ismasked(self):
        """True if the data array has any masked values.

        Deprecated at version TODODASKVER. Use the `is_masked` attribute
        instead.

        **Examples**

        >>> d = cf.Data([[1, 2, 3], [4, 5, 6]])
        >>> print(d.ismasked)
        False
        >>> d[0, ...] = cf.masked
        >>> d.ismasked
        True

        """
        _DEPRECATION_ERROR_ATTRIBUTE(
            self,
            "ismasked",
            message="Use the 'is_masked' attribute instead",
            version="TODODASKVER",
            removed_at="5.0.0",
        )  # pragma: no cover

    @property
    def isscalar(self):
        """True if the data is a 0-d scalar array.

        Deprecated at version TODODASKVER. Use `d.ndim == 0`` instead.

        **Examples**

        >>> d = cf.Data(9, 'm')
        >>> d.isscalar
        True
        >>> d = cf.Data([9], 'm')
        >>> d.isscalar
        False
        >>> d = cf.Data([9, 10], 'm')
        >>> d.isscalar
        False

        """
        _DEPRECATION_ERROR_ATTRIBUTE(
            self,
            "isscalar",
            message="Use 'd.ndim == 0' instead",
            version="TODODASKVER",
            removed_at="5.0.0",
        )  # pragma: no cover

    @property
    def ispartitioned(self):
        """True if the data array is partitioned.

        **Examples**

        >>> d._pmsize
        1
        >>> d.ispartitioned
        False

        >>> d._pmsize
        2
        >>> d.ispartitioned
        False

        """
        _DEPRECATION_ERROR_ATTRIBUTE(
            self, "ispartitioned", version="TODODASKVER", removed_at="5.0.0"
        )  # pragma: no cover

    @property
    def unsafe_array(self):
        """Deprecated at version 3.0.0.

        Use the `array` attribute instead.

        """
        _DEPRECATION_ERROR_ATTRIBUTE(
            self,
            "unsafe_array",
            message="Use the 'array' attribute instead.",
            version="3.0.0",
            removed_at="4.0.0",
        )  # pragma: no cover

    @property
    def varray(self):
        """A numpy array view of the data array.

        Deprecated at version TODODASKVER. Data are now stored as `dask`
        arrays for which, in general, a numpy array view is not
        robust.

        Note that making changes to elements of the returned view changes
        the underlying data.

        .. seealso:: `array`, `to_dask_array`, `datetime_array`

        **Examples**

        >>> a = d.varray
        >>> type(a)
        <type 'numpy.ndarray'>
        >>> a
        array([0, 1, 2, 3, 4])
        >>> a[0] = 999
        >>> d.varray
        array([999, 1, 2, 3, 4])

        """
        _DEPRECATION_ERROR_ATTRIBUTE(
            self,
            "varray",
            message="Data are now stored as `dask` arrays for which, "
            "in general, a numpy array view is not robust.",
            version="TODODASKVER",
            removed_at="5.0.0",
        )  # pragma: no cover

    def expand_dims(self, position=0, i=False):
        """Deprecated at version 3.0.0, use method `insert_dimension`
        instead.

        May get re-instated at a later version.

        """
        _DEPRECATION_ERROR_METHOD(
            self,
            "expand_dims",
            "Use method 'insert_dimension' instead.",
            version="3.0.0",
        )  # pragma: no cover

    def files(self):
        """Deprecated at version 3.4.0, consider using method
        `get_filenames` instead."""
        _DEPRECATION_ERROR_METHOD(
            self,
            "files",
            "Use method `get_filenames` instead.",
            version="3.4.0",
            removed_at="4.0.0",
        )  # pragma: no cover

    def fits_in_one_chunk_in_memory(self, itemsize):
        """Return True if the master array is small enough to be
        retained in memory.

        Deprecated at version TODODASKVER.

        :Parameters:

            itemsize: `int`
                The number of bytes per word of the master data array.

        :Returns:

            `bool`

        **Examples**

        >>> print(d.fits_one_chunk_in_memory(8))
        False

        """
        _DEPRECATION_ERROR_METHOD(
            self,
            "fits_in_one_chunk_in_memory",
            version="TODODASKVER",
            removed_at="5.0.0",
        )  # pragma: no cover

    def close(self):
        """Close all files referenced by the data array.

        Deprecated at version TODODASKVER. All files are now
        automatically closed when not being accessed.

        """
        _DEPRECATION_ERROR_METHOD(
            self,
            "close",
            "All files are now automatically closed when not being accessed.",
            version="TODODASKVER",
            removed_at="5.0.0",
        )  # pragma: no cover

    def chunk(self, chunksize=None, total=None, omit_axes=None, pmshape=None):
        """Partition the data array.

        Deprecated at version TODODASKVER. Use the `rechunk` method
        instead.

        :Parameters:

            chunksize: `int`, optional
                The

            total: sequence of `int`, optional

            omit_axes: sequence of `int`, optional

            pmshape: sequence of `int`, optional

        :Returns:

            `None`

        **Examples**

        >>> d.chunk()
        >>> d.chunk(100000)
        >>> d.chunk(100000, )
        >>> d.chunk(100000, total=[2])
        >>> d.chunk(100000, omit_axes=[3, 4])

        """
        _DEPRECATION_ERROR_METHOD(
            self,
            "chunk",
            message="Use the 'rechunk' method instead.",
            version="TODODASKVER",
            removed_at="5.0.0",
        )  # pragma: no cover

    def dumpd(self):
        """Return a serialisation of the data array.

        Deprecated at version TODODASKVER. Consider inspecting the dask
        array returned by `to_dask_array` instead.

        .. seealso:: `loadd`, `loads`

        :Returns:

            `dict`
                The serialisation.

        **Examples**

        >>> d = cf.Data([[1, 2, 3]], 'm')
        >>> d.dumpd()
        {'Partitions': [{'location': [(0, 1), (0, 3)],
                         'subarray': array([[1, 2, 3]])}],
         'units': 'm',
         '_axes': ['dim0', 'dim1'],
         '_pmshape': (),
         'dtype': dtype('int64'),
         'shape': (1, 3)}

        >>> d.flip(1)
        >>> d.transpose()
        >>> d.Units *= 1000
        >>> d.dumpd()
        {'Partitions': [{'units': 'm',
                         'axes': ['dim0', 'dim1'],
                         'location': [(0, 3), (0, 1)],
                         'subarray': array([[1, 2, 3]])}],
        ` 'units': '1000 m',
         '_axes': ['dim1', 'dim0'],
         '_flip': ['dim1'],
         '_pmshape': (),
         'dtype': dtype('int64'),
         'shape': (3, 1)}

        >>> d.dumpd()
        {'Partitions': [{'units': 'm',
                         'location': [(0, 1), (0, 3)],
                         'subarray': array([[1, 2, 3]])}],
         'units': '10000 m',
         '_axes': ['dim0', 'dim1'],
         '_flip': ['dim1'],
         '_pmshape': (),
         'dtype': dtype('int64'),
         'shape': (1, 3)}

        >>> e = cf.Data(loadd=d.dumpd())
        >>> e.equals(d)
        True

        """
        _DEPRECATION_ERROR_METHOD(
            self,
            "dumpd",
            message="Consider inspecting the dask array returned "
            "by 'to_dask_array' instead.",
            version="TODODASKVER",
            removed_at="5.0.0",
        )  # pragma: no cover

    def dumps(self):
        """Return a JSON string serialisation of the data array.

        Deprecated at version TODODASKVER. Consider inspecting the dask
        array returned by `to_dask_array` instead.

        """
        _DEPRECATION_ERROR_METHOD(
            self,
            "dumps",
            message="Consider inspecting the dask array returned "
            "by 'to_dask_array' instead.",
            version="TODODASKVER",
            removed_at="5.0.0",
        )  # pragma: no cover

    def HDF_chunks(self, *chunks):
        """Get or set HDF chunk sizes.

        The HDF chunk sizes may be used by external code that allows
        `Data` objects to be written to netCDF files.

        Deprecated at version TODODASKVER and is no longer available. Use
        the methods `nc_clear_hdf5_chunksizes`, `nc_hdf5_chunksizes`,
        and `nc_set_hdf5_chunksizes` instead.

        .. seealso:: `nc_clear_hdf5_chunksizes`, `nc_hdf5_chunksizes`,
                     `nc_set_hdf5_chunksizes`

        :Parameters:

            chunks: `dict` or `None`, *optional*
                Specify HDF chunk sizes.

                When no positional argument is provided, the HDF chunk
                sizes are unchanged.

                If `None` then the HDF chunk sizes for each dimension
                are cleared, so that the HDF default chunk size value
                will be used when writing data to disk.

                If a `dict` then it defines for a subset of the
                dimensions, defined by their integer positions, the
                corresponding HDF chunk sizes. The HDF chunk sizes are
                set as a number of elements along the dimension.

        :Returns:

            `dict`
                The HDF chunks for each dimension prior to the change,
                or the current HDF chunks if no new values are
                specified. A value of `None` is an indication that the
                default chunk size should be used for that dimension.

        **Examples**

        >>> d = cf.Data(np.arange(30).reshape(5, 6))
        >>> d.HDF_chunks()
        {0: None, 1: None}
        >>> d.HDF_chunks({1: 2})
        {0: None, 1: None}
        >>> d.HDF_chunks()
        {0: None, 1: 2}
        >>> d.HDF_chunks({1:None})
        {0: None, 1: 2}
        >>> d.HDF_chunks()
        {0: None, 1: None}
        >>> d.HDF_chunks({0: 3, 1: 6})
        {0: None, 1: None}
        >>> d.HDF_chunks()
        {0: 3, 1: 6}
        >>> d.HDF_chunks({1: 4})
        {0: 3, 1: 6}
        >>> d.HDF_chunks()
        {0: 3, 1: 4}
        >>> d.HDF_chunks({1: 999})
        {0: 3, 1: 4}
        >>> d.HDF_chunks()
        {0: 3, 1: 999}
        >>> d.HDF_chunks(None)
        {0: 3, 1: 999}
        >>> d.HDF_chunks()
        {0: None, 1: None}

        """
        _DEPRECATION_ERROR_METHOD(
            self,
            "HDF_chunks",
            message="Use the methods 'nc_clear_hdf5_chunksizes', "
            "'nc_hdf5_chunksizes', and 'nc_set_hdf5_chunksizes' "
            "instead.",
            version="TODODASKVER",
            removed_at="5.0.0",
        )  # pragma: no cover

    def loadd(self, d, chunk=True):
        """Reset the data in place from a dictionary serialisation.

        Deprecated at version TODODASKVER. Consider inspecting the dask
        array returned by `to_dask_array` instead.

        .. seealso:: `dumpd`, `loads`

        :Parameters:

            d: `dict`
                A dictionary serialisation of a `cf.Data` object, such as
                one as returned by the `dumpd` method.

            chunk: `bool`, optional
                If True (the default) then the reset data array will be
                re-partitioned according the current chunk size, as
                defined by the `cf.chunksize` function.

        :Returns:

            `None`

        **Examples**

        >>> d = Data([[1, 2, 3]], 'm')
        >>> e = Data([6, 7, 8, 9], 's')
        >>> e.loadd(d.dumpd())
        >>> e.equals(d)
        True
        >>> e is d
        False

        >>> e = Data(loadd=d.dumpd())
        >>> e.equals(d)
        True

        """
        _DEPRECATION_ERROR_METHOD(
            self,
            "loadd",
            message="Consider inspecting the dask array returned "
            "by 'to_dask_array' instead.",
            version="TODODASKVER",
            removed_at="5.0.0",
        )  # pragma: no cover

    def loads(self, j, chunk=True):
        """Reset the data in place from a string serialisation.

        Deprecated at version TODODASKVER. Consider inspecting the dask
        array returned by `to_dask_array` instead.

        .. seealso:: `dumpd`, `loadd`

        :Parameters:

            j: `str`
                A JSON document string serialisation of a `cf.Data` object.

            chunk: `bool`, optional
                If True (the default) then the reset data array will be
                re-partitioned according the current chunk size, as defined
                by the `cf.chunksize` function.

        :Returns:

            `None`

        """
        _DEPRECATION_ERROR_METHOD(
            self,
            "loads",
            message="Consider inspecting the dask array returned "
            "by 'to_dask_array' instead.",
            version="TODODASKVER",
            removed_at="5.0.0",
        )  # pragma: no cover

    def add_partitions(self, extra_boundaries, pdim):
        """Add partition boundaries.

        Deprecated at version TODODASKVER. Use the `rechunk` method
        instead.

        :Parameters:

            extra_boundaries: `list` of `int`
                The boundaries of the new partitions.

            pdim: `str`
                The name of the axis to have the new partitions.

        :Returns:

            `None`

        **Examples**

        >>> d.add_partitions(    )

        """
        _DEPRECATION_ERROR_METHOD(
            self,
            "add_partitions",
            message="Use the 'rechunk' method instead.",
            version="TODODASKVER",
            removed_at="5.0.0",
        )  # pragma: no cover

    @staticmethod
    def mask_fpe(*arg):
        """Masking of floating-point errors in the results of arithmetic
        operations.

        Deprecated at version TODODASKVER. It is currently not possible
        to control how floating-point errors are handled, due to the
        use of `dask` for handling all array manipulations. This may
        change in the future (see
        https://github.com/dask/dask/issues/3245 for more details).

        If masking is allowed then only floating-point errors which would
        otherwise be raised as `FloatingPointError` exceptions are
        masked. Whether `FloatingPointError` exceptions may be raised is
        determined by `cf.Data.seterr`.

        If called without an argument then the current behaviour is
        returned.

        Note that if the raising of `FloatingPointError` exceptions has
        been suppressed then invalid values in the results of arithmetic
        operations may be subsequently converted to masked values with the
        `mask_invalid` method.

        .. seealso:: `cf.Data.seterr`, `mask_invalid`

        :Parameters:

            arg: `bool`, optional
                The new behaviour. True means that `FloatingPointError`
                exceptions are suppressed and replaced with masked
                values. False means that `FloatingPointError` exceptions
                are raised. The default is not to change the current
                behaviour.

        :Returns:

            `bool`
                The behaviour prior to the change, or the current
                behaviour if no new value was specified.

        **Examples:**

        >>> d = cf.Data([0., 1])
        >>> e = cf.Data([1., 2])

        >>> old = cf.Data.mask_fpe(False)
        >>> old = cf.Data.seterr('raise')
        >>> e/d
        FloatingPointError: divide by zero encountered in divide
        >>> e**123456
        FloatingPointError: overflow encountered in power

        >>> old = cf.Data.mask_fpe(True)
        >>> old = cf.Data.seterr('raise')
        >>> e/d
        <CF Data: [--, 2.0] >
        >>> e**123456
        <CF Data: [1.0, --] >

        >>> old = cf.Data.mask_fpe(True)
        >>> old = cf.Data.seterr('ignore')
        >>> e/d
        <CF Data: [inf, 2.0] >
        >>> e**123456
        <CF Data: [1.0, inf] >

        """
        raise DeprecationError(
            "Data method 'mask_fpe' has been deprecated at version TODODASKVER "
            "and is not available.\n\n"
            "It is currently not possible to control how floating-point errors "
            "are handled, due to the use of `dask` for handling all array "
            "manipulations. This may change in the future (see "
            "https://github.com/dask/dask/issues/3245 for more details)."
        )

    def mask_invalid(self, *args, **kwargs):
        """Mask the array where invalid values occur (NaN or inf).

        Deprecated at version TODODASKVER. Use the method
        `masked_invalid` instead.

        .. seealso:: `where`

        :Parameters:

            {{inplace: `bool`, optional}}

            {{i: deprecated at version 3.0.0}}

        :Returns:

            `Data` or `None`
                The masked data, or `None` if the operation was
                in-place.

        **Examples**

        >>> d = cf.Data([0, 1, 2])
        >>> e = cf.Data([0, 2, 0])
        >>> f = d / e
        >>> f
        <CF Data(3): [nan, 0.5, inf]>
        >>> f.mask_invalid()
        <CF Data(3): [--, 0.5, --]>

        """
        _DEPRECATION_ERROR_METHOD(
            self,
            "mask_invalid",
            message="Use the method 'masked_invalid' instead.",
            version="TODODASKVER",
            removed_at="5.0.0",
        )  # pragma: no cover

    def partition_boundaries(self):
        """Return the partition boundaries for each partition matrix
        dimension.

        Deprecated at version TODODASKVER. Consider using the `chunks`
        attribute instead.

        :Returns:

            `dict`

        **Examples**

        """
        _DEPRECATION_ERROR_METHOD(
            self,
            "partition_boundaries",
            message="Consider using the 'chunks' attribute instead.",
            version="TODODASKVER",
            removed_at="5.0.0",
        )  # pragma: no cover

    def save_to_disk(self, itemsize=None):
        """Deprecated."""
        _DEPRECATION_ERROR_METHOD(
            self, "save_to_disk", removed_at="4.0.0"
        )  # pragma: no cover

    def to_disk(self):
        """Store the data array on disk.

        Deprecated at version TODODASKVER.

        There is no change to partitions whose sub-arrays are already
        on disk.

        :Returns:

            `None`

        **Examples**

        >>> d.to_disk()

        """
        _DEPRECATION_ERROR_METHOD(
            self, "to_disk", version="TODODASKVER", removed_at="5.0.0"
        )  # pragma: no cover

    @staticmethod
    def seterr(all=None, divide=None, over=None, under=None, invalid=None):
        """Set how floating-point errors in the results of arithmetic
        operations are handled.

        Deprecated at version TODODASKVER. It is currently not possible
        to control how floating-point errors are handled, due to the
        use of `dask` for handling all array manipulations. This may
        change in the future (see
        https://github.com/dask/dask/issues/3245 for more details).

        The options for handling floating-point errors are:

        ============  ========================================================
        Treatment     Action
        ============  ========================================================
        ``'ignore'``  Take no action. Allows invalid values to occur in the
                      result data array.

        ``'warn'``    Print a `RuntimeWarning` (via the Python `warnings`
                      module). Allows invalid values to occur in the result
                      data array.

        ``'raise'``   Raise a `FloatingPointError` exception.
        ============  ========================================================

        The different types of floating-point errors are:

        =================  =================================  =================
        Error              Description                        Default treatment
        =================  =================================  =================
        Division by zero   Infinite result obtained from      ``'warn'``
                           finite numbers.

        Overflow           Result too large to be expressed.  ``'warn'``

        Invalid operation  Result is not an expressible       ``'warn'``
                           number, typically indicates that
                           a NaN was produced.

        Underflow          Result so close to zero that some  ``'ignore'``
                           precision was lost.
        =================  =================================  =================

        Note that operations on integer scalar types (such as int16) are
        handled like floating point, and are affected by these settings.

        If called without any arguments then the current behaviour is
        returned.

        .. seealso:: `cf.Data.mask_fpe`, `mask_invalid`

        :Parameters:

            all: `str`, optional
                Set the treatment for all types of floating-point errors
                at once. The default is not to change the current
                behaviour.

            divide: `str`, optional
                Set the treatment for division by zero. The default is not
                to change the current behaviour.

            over: `str`, optional
                Set the treatment for floating-point overflow. The default
                is not to change the current behaviour.

            under: `str`, optional
                Set the treatment for floating-point underflow. The
                default is not to change the current behaviour.

            invalid: `str`, optional
                Set the treatment for invalid floating-point
                operation. The default is not to change the current
                behaviour.

        :Returns:

            `dict`
                The behaviour prior to the change, or the current
                behaviour if no new values are specified.

        **Examples:**

        Set treatment for all types of floating-point errors to
        ``'raise'`` and then reset to the previous behaviours:

        >>> cf.Data.seterr()
        {'divide': 'warn', 'invalid': 'warn', 'over': 'warn', 'under': 'ignore'}
        >>> old = cf.Data.seterr('raise')
        >>> cf.Data.seterr(**old)
        {'divide': 'raise', 'invalid': 'raise', 'over': 'raise', 'under': 'raise'}
        >>> cf.Data.seterr()
        {'divide': 'warn', 'invalid': 'warn', 'over': 'warn', 'under': 'ignore'}

        Set the treatment of division by zero to ``'ignore'`` and overflow
        to ``'warn'`` without changing the treatment of underflow and
        invalid operation:

        >>> cf.Data.seterr(divide='ignore', over='warn')
        {'divide': 'warn', 'invalid': 'warn', 'over': 'warn', 'under': 'ignore'}
        >>> cf.Data.seterr()
        {'divide': 'ignore', 'invalid': 'warn', 'over': 'ignore', 'under': 'ignore'}

        Some examples with data arrays:

        >>> d = cf.Data([0., 1])
        >>> e = cf.Data([1., 2])

        >>> old = cf.Data.seterr('ignore')
        >>> e/d
        <CF Data: [inf, 2.0] >
        >>> e**12345
        <CF Data: [1.0, inf] >

        >>> cf.Data.seterr(divide='warn')
        {'divide': 'ignore', 'invalid': 'ignore', 'over': 'ignore', 'under': 'ignore'}
        >>> e/d
        RuntimeWarning: divide by zero encountered in divide
        <CF Data: [inf, 2.0] >
        >>> e**12345
        <CF Data: [1.0, inf] >

        >>> old = cf.Data.mask_fpe(False)
        >>> cf.Data.seterr(over='raise')
        {'divide': 'warn', 'invalid': 'ignore', 'over': 'ignore', 'under': 'ignore'}
        >>> e/d
        RuntimeWarning: divide by zero encountered in divide
        <CF Data: [inf, 2.0] >
        >>> e**12345
        FloatingPointError: overflow encountered in power

        >>> cf.Data.mask_fpe(True)
        False
        >>> cf.Data.seterr(divide='ignore')
        {'divide': 'warn', 'invalid': 'ignore', 'over': 'raise', 'under': 'ignore'}
        >>> e/d
        <CF Data: [inf, 2.0] >
        >>> e**12345
        <CF Data: [1.0, --] >

        """
        raise DeprecationError(
            "Data method 'seterr' has been deprecated at version TODODASKVER "
            "and is not available.\n\n"
            "It is currently not possible to control how floating-point errors "
            "are handled, due to the use of `dask` for handling all array "
            "manipulations. This may change in the future (see "
            "https://github.com/dask/dask/issues/3245 for more details)."
        )

    @classmethod
    def reconstruct_sectioned_data(cls, sections, cyclic=(), hardmask=None):
        """Expects a dictionary of Data objects with ordering
        information as keys, as output by the section method when called
        with a Data object. Returns a reconstructed cf.Data object with
        the sections in the original order.

        Deprecated at version TODODASKVER and is no longer available.

        :Parameters:

            sections: `dict`
                The dictionary of `Data` objects with ordering information
                as keys.

        :Returns:

            `Data`
                The resulting reconstructed Data object.

        **Examples**

        >>> d = cf.Data(numpy.arange(120).reshape(2, 3, 4, 5))
        >>> x = d.section([1, 3])
        >>> len(x)
        8
        >>> e = cf.Data.reconstruct_sectioned_data(x)
        >>> e.equals(d)
        True

        """
        raise DeprecationError(
            "Data method 'reconstruct_sectioned_data' has been deprecated "
            "at version TODODASKVER and is no longer available"
        )

    @classmethod
    def concatenate_data(cls, data_list, axis):
        """Concatenates a list of Data objects along the specified axis.

        See cf.Data.concatenate for details.

        In the case that the list contains only one element, that element
        is simply returned.

        :Parameters:

            data_list: `list`
                The list of data objects to concatenate.

            axis: `int`
                The axis along which to perform the concatenation.

        :Returns:

            `Data`
                The resulting single `Data` object.

        """
        raise DeprecationError(
            "Data method 'concatenate_data' has been deprecated at "
            "version TODODASKVER and is no longer available. Use "
            "'concatenate' instead."
        )



import datetime
import faulthandler
import os
import re
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cf


class FieldTest(unittest.TestCase):
    filename2 = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "test_file2.nc"
    )

    f2 = cf.read(filename2)

    x = cf.example_field(1)

    test_only = []

    def test_FieldList(self):
        cf.FieldList(self.x)
        cf.FieldList([self.x])

    def test_FieldList__add__iadd__(self):
        f = cf.FieldList(self.x)

        f = f + f.copy()
        self.assertEqual(len(f), 2)
        self.assertIsInstance(f, cf.FieldList)

        f += (f[0].copy(),)
        self.assertEqual(len(f), 3)

        f += [f[0].copy()]
        self.assertEqual(len(f), 4)
        self.assertIsInstance(f, cf.FieldList)

        f += list(f[0].copy())
        self.assertEqual(len(f), 5)

        f += f.copy()
        self.assertEqual(len(f), 10)

        f = f + f.copy()
        self.assertEqual(len(f), 20)

    def test_FieldList__contains__(self):
        f = cf.FieldList(self.x)

        f.append(self.x.copy())
        f[1] *= 10
        g = self.x * 10
        self.assertIn(g, f)
        self.assertNotIn(34.6, f)

    def test_FieldList__len__(self):
        f = cf.FieldList(self.x)

        self.assertEqual(len(cf.FieldList()), 0)
        self.assertEqual(len(f), 1)
        f.append(f[0])
        self.assertEqual(len(f), 2)
        f.extend(f)
        self.assertEqual(len(f), 4)

    def test_FieldList__mul__imul__(self):
        f = cf.FieldList()
        f = f * 4
        self.assertEqual(len(f), 0)
        self.assertIsInstance(f, cf.FieldList)

        f = cf.FieldList()
        f *= 4
        self.assertEqual(len(f), 0)
        self.assertIsInstance(f, cf.FieldList)

        f = cf.FieldList(self.x) * 4
        self.assertEqual(len(f), 4)
        self.assertIsInstance(f, cf.FieldList)

        f = cf.FieldList(self.x)
        f *= 4
        self.assertEqual(len(f), 4)
        self.assertIsInstance(f, cf.FieldList)

        f = f * 2
        self.assertEqual(len(f), 8)
        self.assertIsInstance(f, cf.FieldList)

        f *= 3
        self.assertEqual(len(f), 24)
        self.assertIsInstance(f, cf.FieldList)

    def test_FieldList__repr__(self):
        f = cf.FieldList(self.x)
        f += f
        repr(f)

    def test_FieldList_append_extend(self):
        # Append
        f = cf.FieldList()

        f.append(self.x)
        self.assertEqual(len(f), 1)
        self.assertIsInstance(f, cf.FieldList)

        f.append(f[0])
        self.assertEqual(len(f), 2)
        self.assertIsInstance(f, cf.FieldList)

        f.append(f[0])
        self.assertEqual(len(f), 3)

        # Extend
        f = cf.FieldList()

        f.extend([self.x])
        self.assertEqual(len(f), 1)
        self.assertIsInstance(f, cf.FieldList)

        f.extend(f)
        self.assertEqual(len(f), 2)
        self.assertIsInstance(f, cf.FieldList)

        f.extend(f)
        self.assertEqual(len(f), 4)

    def test_FieldList_copy(self):
        f = cf.FieldList(self.x)

        f.append(f[0].copy())
        g = f.copy()
        self.assertTrue(f.equals(f, verbose=2))
        self.assertTrue(f.equals(g, verbose=2))

    def test_FieldList__getslice__(self):
        f = cf.FieldList(self.x)

        f.append(f[0])

        f[0:1]
        f[1:2]
        f[:1]
        f[1:]

    def test_FieldList_count(self):
        f = cf.FieldList(self.x)

        self.assertEqual(f.count(f[0]), 1)

        f = f * 7
        self.assertEqual(f.count(f[0]), 7)

        f[3] = f[0] * 99
        f[5] = f[0] * 99
        self.assertEqual(f.count(f[0]), 5)
        self.assertEqual(f.count(f[3]), 2)

    def test_FieldList_equals(self):
        f = cf.FieldList(self.x)

        g = f.copy()
        self.assertTrue(f.equals(f, verbose=2))
        self.assertTrue(f.equals(g, verbose=2))

        f += g.copy()
        self.assertTrue(f.equals(f, verbose=2))
        self.assertEqual(len(f), 2)
        g = f.copy()
        self.assertTrue(f.equals(g, verbose=2))
        self.assertTrue(f.equals(g, unordered=True))

        h = self.f2.copy()
        for x in h:
            x.standard_name = "eastward_wind"

        f.extend(h)
        self.assertTrue(f.equals(f, verbose=2))
        self.assertTrue(f.equals(f.copy()))

        g = f.copy()[::-1]
        self.assertFalse(f.equals(g))
        self.assertTrue(f.equals(g, unordered=True))

        g = g[:-1]
        self.assertFalse(f.equals(g))
        self.assertFalse(f.equals(g, unordered=True))

        g.append(g[0])
        self.assertFalse(f.equals(g))
        self.assertFalse(f.equals(g, unordered=True))

        h *= 3
        self.assertFalse(f.equals(h))
        self.assertFalse(f.equals(h, unordered=True))

    def test_FieldList_insert_pop_remove(self):
        # Insert
        f = cf.FieldList(self.x)
        g = f[0].copy()

        f.insert(0, g)
        self.assertEqual(len(f), 2)
        self.assertIsInstance(f, cf.FieldList)

        g = g + 10
        f.insert(-1, g)
        self.assertEqual(len(f), 3)
        self.assertEqual(f[0].maximum(), (f[1].maximum() - 10))
        self.assertIsInstance(f, cf.FieldList)

        # Pop
        f = cf.FieldList(self.x)
        g = f[0]
        h = f[0] + 10
        f.append(h)

        z = f.pop(0)
        self.assertIs(z, g)
        self.assertEqual(len(f), 1)
        self.assertIsInstance(f, cf.FieldList)

        z = f.pop(-1)
        self.assertIs(z, h)
        self.assertEqual(len(f), 0)
        self.assertIsInstance(f, cf.FieldList)

        # Remove
        f = cf.FieldList(self.x)
        g = f[0] + 10

        f.append(g)
        self.assertEqual(len(f), 2)

        f.remove(g)
        self.assertEqual(len(f), 1)
        self.assertIsInstance(f, cf.FieldList)

        with self.assertRaises(Exception):
            f.remove(f[0] * -99)

        f.remove(f[0].copy())
        self.assertEqual(len(f), 0)
        self.assertIsInstance(f, cf.FieldList)

    def test_FieldList_reverse(self):
        f = cf.FieldList(self.x)

        g = f[0]
        h = f[0] + 10
        f.append(h)

        self.assertIs(g, f[0])
        self.assertIs(h, f[1])

        f.reverse()
        self.assertIsInstance(f, cf.FieldList)
        self.assertEqual(len(f), 2)
        self.assertIs(g, f[1])
        self.assertIs(h, f[0])

    def test_FieldList_select(self):
        f = cf.FieldList(self.x)

        g = f("not this one")
        self.assertIsInstance(g, cf.FieldList)
        self.assertEqual(len(g), 0)

        g = f("air_temperature")
        self.assertIsInstance(g, cf.FieldList)
        self.assertEqual(len(g), 1, len(g))

        g = f(re.compile("^air"))
        self.assertIsInstance(g, cf.FieldList)
        self.assertEqual(len(g), 1, len(g))

        f *= 9
        f[4] = f[0].copy()
        f[4].standard_name = "this one"
        f[6] = f[0].copy()
        f[6].standard_name = "this one"

        g = f(re.compile("^air"))
        self.assertIsInstance(g, cf.FieldList)
        self.assertEqual(len(g), 7, len(g))

        g = f("this one")
        self.assertIsInstance(g, cf.FieldList)
        self.assertEqual(len(g), 2)

        # select_by_Units
        f[1] = f[1].override_units(cf.Units("watt"))
        f[3] = f[3].override_units(cf.Units("K @ 273.15"))

        self.assertEqual(len(f.select_by_units()), 9)
        self.assertEqual(len(f.select_by_units(cf.Units("K"))), 7)
        self.assertEqual(len(f.select_by_units(cf.Units("K"), exact=False)), 8)
        self.assertEqual(len(f.select_by_units("K")), 7)
        self.assertEqual(len(f.select_by_units("K", exact=False)), 8)
        self.assertEqual(len(f.select_by_units(re.compile("^K @|watt"))), 2)

        self.assertEqual(len(f.select_by_units("long_name=qwery:asd")), 0)

        # select_by_ncvar
        for a in f:
            a.nc_set_variable("ta")

        f[1].nc_set_variable("qwerty")
        f[4].nc_set_variable("ta2")

        self.assertEqual(len(f.select_by_ncvar()), 9)
        self.assertEqual(len(f.select_by_ncvar("qwerty")), 1)
        self.assertEqual(len(f.select_by_ncvar("ta")), 7)
        self.assertEqual(len(f.select_by_ncvar("ta2")), 1)
        self.assertEqual(len(f.select_by_ncvar(re.compile("^ta"))), 8)

    def test_FieldList_select_by_construct(self):
        x = self.x.copy()
        x.del_construct("time")

        f = cf.FieldList(x)
        f.extend(self.f2.copy())

        g = f.select_by_construct()
        self.assertEqual(len(g), 2)

        g = f.select_by_construct("latitude")
        self.assertEqual(len(g), 2)

        g = f.select_by_construct("latitude", "longitude")
        self.assertEqual(len(g), 2)

        g = f.select_by_construct("latitude", "time")
        self.assertEqual(len(g), 1)

        g = f.select_by_construct("latitude", "time", OR=False)
        self.assertEqual(len(g), 1)

        g = f.select_by_construct("latitude", "time", OR=True)
        self.assertEqual(len(g), 2)

        g = f.select_by_construct(longitude=cf.gt(0))
        self.assertEqual(len(g), 2)

        g = f.select_by_construct(
            longitude=cf.gt(0), time=cf.le(cf.dt("2008-12-01"))
        )
        self.assertEqual(len(g), 1)

        g = f.select_by_construct(
            longitude=cf.gt(0), time=cf.le(cf.dt("2008-12-01")), OR=True
        )
        self.assertEqual(len(g), 2)

        g = f.select_by_construct(
            "longitude",
            longitude=cf.gt(0),
            time=cf.le(cf.dt("2008-12-01")),
            OR=True,
        )
        self.assertEqual(len(g), 2)

        g = f.select_by_construct(
            "latitude",
            longitude=cf.gt(0),
            time=cf.le(cf.dt("2008-12-01")),
            OR=True,
        )
        self.assertEqual(len(g), 2)

        g = f.select_by_construct(
            "time",
            longitude=cf.gt(0),
            time=cf.le(cf.dt("2008-12-01")),
            OR=True,
        )
        self.assertEqual(len(g), 2)

        g = f.select_by_construct(
            "time",
            longitude=cf.gt(0),
            time=cf.le(cf.dt("2008-12-01")),
            OR=False,
        )
        self.assertEqual(len(g), 1)

        g = f.select_by_construct(
            "time",
            longitude=cf.gt(0),
            time=cf.le(cf.dt("2008-12-01")),
            OR=False,
        )
        self.assertEqual(len(g), 1)

        g = f.select_by_construct("qwerty")
        self.assertEqual(len(g), 0)

        g = f.select_by_construct("qwerty", "latitude")
        self.assertEqual(len(g), 0)

        g = f.select_by_construct("qwerty", "latitude", OR=True)
        self.assertEqual(len(g), 2)

        g = f.select_by_construct("qwerty", "time", "longitude")
        self.assertEqual(len(g), 0)

        g = f.select_by_construct("qwerty", "time", "longitude", OR=True)
        self.assertEqual(len(g), 2)

        g = f.select_by_construct(longitude=cf.gt(7.6))
        self.assertEqual(len(g), 1)

        g = f.select_by_construct(
            longitude=cf.gt(0), time=cf.le(cf.dt("1999-12-01"))
        )
        self.assertEqual(len(g), 0)

    def test_FieldList_select_field(self):
        f = cf.FieldList(self.x)

        with self.assertRaises(Exception):
            f.select_field("not this one")

        self.assertIsNone(f.select_field("not this one", default=None))

        g = f.select_field("air_temperature")
        self.assertIsInstance(g, cf.Field)

        g = f.select_field(re.compile("^air_temp"))
        self.assertIsInstance(g, cf.Field)

        with self.assertRaises(Exception):
            g = f.select_field(re.compile("^QWERTY"))

    def test_FieldList_concatenate(self):
        f = self.f2[0]

        g = cf.FieldList([f[0], f[1:456], f[456:]])

        h = g.concatenate(axis=0)
        self.assertTrue(f.equals(h, verbose=2))

        h = g.concatenate(axis=0, cull_graph=False)
        self.assertTrue(f.equals(h, verbose=2))

    def test_FieldList_index(self):
        f = self.f2[0]

        a, b, c = [f[0], f[1:456], f[456:]]
        g = cf.FieldList([a, b, c])

        self.assertEqual(g.index(a), 0)
        self.assertEqual(g.index(a, start=0), 0)
        self.assertEqual(g.index(a, stop=1), 0)
        self.assertEqual(g.index(a, stop=-2), 0)
        self.assertEqual(g.index(a, stop=2), 0)
        self.assertEqual(g.index(b), 1)
        self.assertEqual(g.index(b, start=0), 1)
        self.assertEqual(g.index(b, start=1, stop=2), 1)
        self.assertEqual(g.index(c), 2)
        self.assertEqual(g.index(c, start=0), 2)
        self.assertEqual(g.index(c, start=1), 2)
        self.assertEqual(g.index(c, start=2), 2)
        self.assertEqual(g.index(c, start=-1), 2)

        with self.assertRaises(Exception):
            g.index(f)

        with self.assertRaises(Exception):
            g.index(a, start=1)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import datetime
import faulthandler
import os
import stat
import subprocess
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cf


class cfaTest(unittest.TestCase):
    def setUp(self):
        self.test_file = "cfa_test.sh"
        self.test_path = os.path.join(os.getcwd(), self.test_file)

        # Need ./cfa_test.sh to be made executable to run it. Locally that
        # may already be true but for testing dists etc. must chmod here:
        os.chmod(
            self.test_path,
            stat.S_IRUSR
            | stat.S_IROTH
            | stat.S_IRGRP
            | stat.S_IXUSR  # reading
            | stat.S_IXOTH
            | stat.S_IXGRP  # executing
            | stat.S_IWUSR,  # writing
        )

    def test_cfa(self):
        # In the script, STDERR from cfa commands is redirected to
        # (overwrite) its STDOUT, so Popen's stdout is really the cfa
        # commands' stderr:
        cfa_test = subprocess.Popen(
            ["./" + self.test_file],
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
        )
        cfa_stderr_via_stdout_channel, _ = cfa_test.communicate("yes\n")
        returncode = cfa_test.returncode
        if returncode != 0:
            self.fail(
                f"A cfa command failed (see script's 'exit {returncode}' "
                "point) with error:\n"
                f"{cfa_stderr_via_stdout_channel.decode('utf-8')}"
            )
        # else: (passes by default)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import cfdm

from . import mixin


class TiePointIndex(mixin.PropertiesData, cfdm.TiePointIndex):
    """A tie point index variable with properties.

    Space may be saved by storing a subsample of the coordinates. The
    uncompressed coordinates can be reconstituted by interpolation
    from the subsampled coordinate values, also called either "tie
    points" or "bounds tie points".

    For each interpolated dimension, the locations of the (bounds) tie
    points are defined by a corresponding tie point index variable,
    which also indicates the extent of each continuous area.

    **NetCDF interface**

    {{netCDF variable}}

    The netCDF subsampled dimension name may be accessed with the
    `nc_set_subsampled_dimension`, `nc_get_subsampled_dimension`,
    `nc_del_subsampled_dimension` and `nc_has_subsampled_dimension`
    methods.

    The netCDF subsampled dimension group structure may be accessed
    with the `nc_set_subsampled_dimension`,
    `nc_get_subsampled_dimension`, `nc_subsampled_dimension_groups`,
    `nc_clear_subsampled_dimension_groups`, and
    `nc_set_subsampled_dimension_groups` methods.

    The netCDF interpolation subarea dimension name may be accessed
    with the `nc_set_interpolation_subarea_dimension`,
    `nc_get_interpolation_subarea_dimension`,
    `nc_del_interpolation_subarea_dimension`, and
    `nc_has_interpolation_subarea_dimension` methods.

    The netCDF interpolation subarea dimension group structure may be
    accessed with the `nc_set_interpolation_subarea_dimension`,
    `nc_get_interpolation_subarea_dimension`,
    `nc_interpolation_subarea_dimension_groups`,
    `nc_clear_interpolation_subarea_dimension_groups` and
    `nc_set_interpolation_subarea_dimension_groups` methods.

    .. versionadded:: 3.14

    .. seealso:: `InterpolationParameter`

    """

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        """
        return super().__repr__().replace("<", "<CF ", 1)



from . import CompressedArrayMixin


class RaggedArrayMixin(CompressedArrayMixin):
    """Mixin class for compressed ragged arrays.

    .. versionadded:: TODODASKVER

    """

    def to_dask_array(self, chunks="auto"):
        """Convert the data to a `dask` array.

        .. versionadded:: TODODASKVER

        :Parameters:

            chunks: `int`, `tuple`, `dict` or `str`, optional
                Specify the chunking of the returned dask array.

                Any value accepted by the *chunks* parameter of the
                `dask.array.from_array` function is allowed.

                The chunk sizes implied by *chunks* for a dimension that
                has been fragmented are ignored and replaced with values
                that are implied by that dimensions fragment sizes.

        :Returns:

            `dask.array.Array`
                The `dask` array representation.

        """
        from functools import partial

        import dask.array as da
        from dask import config
        from dask.array.core import getter, normalize_chunks
        from dask.base import tokenize

        name = (f"{self.__class__.__name__}-{tokenize(self)}",)

        dtype = self.dtype

        context = partial(config.set, scheduler="synchronous")

        compressed_dimensions = self.compressed_dimensions()
        conformed_data = self.conformed_data()
        compressed_data = conformed_data["data"]

        # If possible, convert the compressed data to a dask array
        # that doesn't support concurrent reads. This prevents
        # "compute called by compute" failures problems at compute
        # time.
        #
        # TODO: This won't be necessary if this is refactored so that
        #       the compressed data is part of the same dask graph as
        #       the compressed subarrays.
        compressed_data = self._lock_file_read(compressed_data)

        # Get the (cfdm) subarray class
        Subarray = self.get_Subarray()
        subarray_name = Subarray().__class__.__name__

        # Set the chunk sizes for the dask array
        chunks = self.subarray_shapes(chunks)
        chunks = normalize_chunks(
            self.subarray_shapes(chunks),
            shape=self.shape,
            dtype=dtype,
        )

        dsk = {}
        for u_indices, u_shape, c_indices, chunk_location in zip(
            *self.subarrays(chunks)
        ):
            subarray = Subarray(
                data=compressed_data,
                indices=c_indices,
                shape=u_shape,
                compressed_dimensions=compressed_dimensions,
                context_manager=context,
            )

            key = f"{subarray_name}-{tokenize(subarray)}"
            dsk[key] = subarray

            dsk[name + chunk_location] = (
                getter,
                key,
                Ellipsis,
                False,
                False,
            )

        # Return the dask array
        return da.Array(dsk, name[0], chunks=chunks, dtype=dtype)



import sys

import numpy as np


def cmp(a, b):
    """Workaround to get a Python-2-like `cmp` function in Python 3."""
    return (a > b) - (a < b)


_codes = {
    1: ("x", float),
    2: ("y", float),
    3: ("y_domain_lower_bound", float),
    4: ("x_domain_lower_bound", float),
    5: ("y_domain_upper_bound", float),
    6: ("x_domain_upper_bound", float),
    7: ("z_domain_lower_bound", float),
    8: ("z_domain_upper_bound", float),
    10: ("title", str),
    11: ("domain_title", str),
    12: ("x_lower_bound", float),
    13: ("x_upper_bound", float),
    14: ("y_lower_bound", float),
    15: ("y_upper_bound", float),
}


class ExtraData(dict):
    """Extends dictionary class with a comparison method between extra
    data for different records."""

    _key_to_type = dict([(key, typ) for key, typ in _codes.values()])

    def sorted_keys(self):
        k = self.keys()
        k.sort()
        return k

    _tolerances = {np.dtype(np.float32): 1e-5, np.dtype(np.float64): 1e-13}

    def _cmp_floats(self, a, b, tolerance):
        if a == b:
            return 0

        delta = abs(b * tolerance)
        if a < b - delta:
            return -1

        if a > b + delta:
            return 1

        return 0

    def _cmp_float_arrays(self, avals, bvals):
        n = len(avals)
        c = cmp(n, len(bvals))
        if c != 0:
            return c

        tolerance = self._tolerances[avals.dtype]
        for i in range(n):
            c = self._cmp_floats(avals[i], bvals[i], tolerance)
            if c != 0:
                return c

        return 0

    def __cmp__(self, other):
        """Compare two extra data dictionaries returned by unpacker."""
        if other is None:
            return 1
        ka = self.sorted_keys()
        kb = other.sorted_keys()
        c = cmp(ka, kb)
        if c != 0:
            return c

        for key in ka:
            valsa = self[key]
            valsb = other[key]
            typ = self._key_to_type[key]
            if typ == float:
                c = self._cmp_float_arrays(valsa, valsb)
            elif type == str:
                c = cmp(valsa, valsb)
            else:
                assert False

            if c != 0:
                return c

        return 0


class ExtraDataUnpacker:

    _int_types = {4: np.int32, 8: np.int64}
    _float_types = {4: np.float32, 8: np.float64}

    def __init__(self, raw_extra_data, word_size, byte_ordering):
        self.rdata = raw_extra_data
        self.ws = word_size
        self.itype = self._int_types[word_size]
        self.ftype = self._float_types[word_size]
        # byte_ordering is 'little_endian' or 'big_endian'
        # sys.byteorder is 'little' or 'big'
        self.is_swapped = not byte_ordering.startswith(sys.byteorder)

    def next_words(self, n):
        """return next n words as raw data string, and pop them off the
        front of the string."""
        pos = n * self.ws
        rv = self.rdata[:pos]
        assert len(rv) == pos
        self.rdata = self.rdata[pos:]
        return rv

    def convert_bytes_to_string(self, st):
        """Convert bytes to string.

        :Returns:

            `str`

        """
        if self.is_swapped:
            indices = slice(None, None, -1)
        else:
            indices = slice(None)

        st = "".join(
            [
                st[pos : pos + self.ws][indices].decode("utf-8")
                for pos in range(0, len(st), self.ws)
            ]
        )

        while st.endswith("\x00"):
            st = st[:-1]

        return st

    def get_data(self):
        """Get the (key, value) pairs for extra data.

        :Returns:

            `ExtraData`

        """
        d = {}
        while self.rdata:
            i = np.frombuffer(self.next_words(1), self.itype)[0]
            if i == 0:
                break

            ia, ib = divmod(i, 1000)
            key, etype = _codes[ib]

            rawvals = self.next_words(ia)
            if etype == float:
                vals = np.frombuffer(rawvals, self.ftype)
            elif etype == str:
                vals = np.array([self.convert_bytes_to_string(rawvals)])

            if key not in d:
                d[key] = vals
            else:
                d[key] = np.append(d[key], vals)

        return ExtraData(d)



import numpy as np


class FileArrayMixin:
    """Mixin class for an array stored in a file.

    .. versionadded:: TODODASKVER

    """

    @property
    def _dask_meta(self):
        """The metadata for the containing dask array.

        This is the kind of array that will result from slicing the
        file array.

        .. versionadded:: TODODASKVER

        .. seealso:: `dask.array.from_array`

        """
        return np.array((), dtype=self.dtype)
